# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import numpy
import math
import scipy.stats
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.data import FecDataView

logger = FormatAdapter(logging.getLogger(__name__))

# Define profiler time scale in ms
_MS_SCALE = (1.0 / 200000.0)

# Define the positions in tags
_START_TIME = 0
_DURATION = 1


class ProfileData(object):
    """
    A container for profile data.
    """

    START_TIME = _START_TIME
    DURATION = _DURATION

    __slots__ = (
        # A dictionary of tag label to numpy array of start times and durations
        "_tags",

        # A list of tag labels indexed by the tag ID
        "_tag_labels",

        # The maximum time recorded
        "_max_time"
    )

    def __init__(self, tag_labels):
        """
        :param list(str) tag_labels: A list of labels indexed by tag ID
        """
        self._tag_labels = tag_labels
        self._tags = dict()
        self._max_time = None

    def add_data(self, data):
        """
        Add profiling data read from the profile section.

        :param bytearray data:
            Data read from the profile section on the machine
        """
        samples = numpy.asarray(data, dtype="uint8").view(dtype="<u4")

        # Slice data to separate times, tags and flags
        sample_times = samples[::2]
        sample_tags_and_flags = samples[1::2]

        # Further split the tags and flags word into separate arrays of tags
        # and flags
        sample_tags = numpy.bitwise_and(sample_tags_and_flags, 0x7FFFFFFF)
        sample_flags = numpy.right_shift(sample_tags_and_flags, 31)

        # Find indices of samples relating to entries and exits
        sample_entry_indices = numpy.where(sample_flags == 1)
        sample_exit_indices = numpy.where(sample_flags == 0)

        # Convert count-down times to count up times from 1st sample
        sample_times_ms = numpy.multiply(
            numpy.subtract(sample_times[0], sample_times),
            _MS_SCALE, dtype=float)

        # Slice tags and times into entry and exits
        entry_tags = sample_tags[sample_entry_indices]
        entry_times_ms = sample_times_ms[sample_entry_indices]
        exit_tags = sample_tags[sample_exit_indices]
        exit_times_ms = sample_times_ms[sample_exit_indices]

        # Loop through unique tags
        for tag in numpy.unique(sample_tags):
            self._add_tag_data(
                entry_tags, entry_times_ms, exit_tags, exit_times_ms, tag)

    def _add_tag_data(
            self, entry_tags, entry_times, exit_tags, exit_times, tag):
        """
        :param ~numpy.ndarray entry_tags:
        :param ~numpy.ndarray entry_times:
        :param ~numpy.ndarray exit_tags:
        :param ~numpy.ndarray exit_times:
        :param int tag:
        """
        # pylint: disable=too-many-arguments
        tag_label = self._tag_labels.get(tag, None)
        if tag_label is None:
            logger.warning("Unknown tag {} in profile data", tag)
            tag_label = "UNKNOWN"

        # Get indices where these tags occur
        tag_entry_indices = numpy.where(entry_tags == tag)
        tag_exit_indices = numpy.where(exit_tags == tag)

        # Use these to get subset for this tag
        tag_entry_times = entry_times[tag_entry_indices]
        tag_exit_times = exit_times[tag_exit_indices]

        # If the first exit is before the first
        # Entry, add a dummy entry at beginning
        if tag_exit_times[0] < tag_entry_times[0]:
            logger.warning("Profile starts mid-tag")
            tag_entry_times = numpy.append(0.0, tag_entry_times)

        if len(tag_entry_times) > len(tag_exit_times):
            logger.warning("profile finishes mid-tag")
            tag_entry_times = tag_entry_times[
                :len(tag_exit_times) - len(tag_entry_times)]

        # Subtract entry times from exit times to get durations of each
        # call in ms
        tag_durations = numpy.subtract(tag_exit_times, tag_entry_times)

        # Add entry times and durations to dictionary
        self._tags[tag_label] = (tag_entry_times, tag_durations)

        # Keep track of the maximum time
        self._max_time = numpy.max(tag_entry_times + tag_durations)

    @property
    def tags(self):
        """
        The tags recorded as labels.

        :rtype: list(str)
        """
        return self._tags.keys()

    def get_mean_ms(self, tag):
        """
        Get the mean time in milliseconds spent on operations with the
        given tag.

        :param str tag: The tag to get the mean time for
        :rtype: float
        """
        return numpy.average(self._tags[tag][_DURATION])

    def get_n_calls(self, tag):
        """
        Get the number of times the given tag was recorded.

        :param str tag: The tag to get the number of calls of
        :rtype: int
        """
        return self._tags[tag][_DURATION].size

    def get_mean_n_calls_per_ts(self, tag):
        """
        Get the mean number of times the given tag was recorded per timestep.

        :param str tag: The tag to get the data for
        :rtype: float
        """
        time_step_ms = FecDataView.get_simulation_time_step_ms()
        n_points = math.ceil(
            self._max_time / time_step_ms)
        endpoint = n_points * time_step_ms
        bins = numpy.linspace(0, endpoint, n_points + 1)
        return numpy.average(numpy.histogram(
            self._tags[tag][_START_TIME], bins)[0])

    def get_mean_ms_per_ts(self, tag):
        """
        Get the mean time in milliseconds spent on operations with the
        given tag per timestep.

        :param str tag: The tag to get the data for
        :rtype: float
        """
        time_step_ms = FecDataView.get_simulation_time_step_ms()
        n_points = math.ceil(
            self._max_time / time_step_ms)
        endpoint = n_points * time_step_ms
        bins = numpy.linspace(0, endpoint, n_points + 1)
        mean_per_ts = scipy.stats.binned_statistic(
            self._tags[tag][_START_TIME], self._tags[tag][_DURATION],
            "mean", bins).statistic
        mean_per_ts[numpy.isnan(mean_per_ts)] = 0
        return numpy.average(
            mean_per_ts[numpy.logical_not(numpy.isnan(mean_per_ts))])



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.profiling import AbstractHasProfileData


_FMT_A = "{: <{}s} {: <7s} {: <14s} {: <14s} {: <14s}\n"
_FMT_B = "{:-<{}s} {:-<7s} {:-<14s} {:-<14s} {:-<14s}\n"
_FMT_C = "{: <{}s} {: >7d} {: >14.6f} {: >14.6f} {: >14.6f}\n"


def profile_data_gatherer():
    """
    Gets all the profiling data recorded by vertices and writes it to files.
    """
    progress = ProgressBar(
        FecDataView.get_n_placements(), "Getting profile data")
    provenance_file_path = FecDataView.get_app_provenance_dir_path()

    # retrieve provenance data from any cores that provide data
    for placement in progress.over(FecDataView.iterate_placemements()):
        if isinstance(placement.vertex, AbstractHasProfileData):
            # get data
            profile_data = placement.vertex.get_profile_data(placement)
            if profile_data.tags:
                _write(placement, profile_data, provenance_file_path)


def _write(p, profile_data, directory):
    """
    :param ~.Placement p:
    :param ProfileData profile_data:
    :param str directory:
    """
    max_tag_len = max(len(tag) for tag in profile_data.tags)
    file_name = os.path.join(
        directory, f"{p.x}_{p.y}_{ p.p}_profile.txt")

    # write profile data to file, creating if necessary
    with open(file_name, "a", encoding="utf-8") as f:
        # Write header
        f.write(_FMT_A.format(
            "tag", max_tag_len, "n_calls", "mean_ms", "n_calls_per_ts",
            "mean_ms_per_ts"))
        f.write(_FMT_B.format("", max_tag_len, "", "", "", ""))

        # Write content
        for tag in profile_data.tags:
            f.write(_FMT_C.format(
                tag, max_tag_len, profile_data.get_n_calls(tag),
                profile_data.get_mean_ms(tag),
                profile_data.get_mean_n_calls_per_ts(tag),
                profile_data.get_mean_ms_per_ts(tag)))



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.config_holder import get_config_int
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.placements import Placement
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utility_models import (
    ChipPowerMonitorMachineVertex)

_LABEL = "chip_power_monitor_{}_vertex_for_chip({}:{})"


def insert_chip_power_monitors_to_graphs(placements):
    """
    Adds chip power monitors into a given graph.

    param Placements placements:
    """
    sampling_frequency = get_config_int("EnergyMonitor", "sampling_frequency")
    machine = FecDataView.get_machine()
    # create progress bar
    progress = ProgressBar(
        machine.n_chips, "Adding Chip power monitors to Graph")

    for chip in progress.over(machine.chips):
        vertex = ChipPowerMonitorMachineVertex(
            f"ChipPowerMonitor on {chip.x}, {chip.y}",
            sampling_frequency=sampling_frequency)
        cores = __cores(machine, chip.x, chip.y)
        p = cores[placements.n_placements_on_chip(chip.x, chip.y)]
        placements.add_placement(Placement(vertex, chip.x, chip.y, p))


def __cores(machine, x, y):
    return [p.processor_id for p in machine.get_chip_at(x, y).processors
            if not p.is_monitor]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import logging
from enum import IntEnum
import numpy
from spinn_utilities.config_holder import get_config_int
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from data_specification.enums import DataType
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import VariableSDRAM
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.buffer_management import (
    recording_utilities)
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    AbstractReceiveBuffersToHost)
from spinn_front_end_common.interface.provenance import ProvenanceWriter
from spinn_front_end_common.utilities.constants import (
    SYSTEM_BYTES_REQUIREMENT, SIMULATION_N_BYTES, BYTES_PER_WORD)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.interface.simulation.simulation_utilities import (
    get_simulation_header_array)

logger = FormatAdapter(logging.getLogger(__name__))
BINARY_FILE_NAME = "chip_power_monitor.aplx"
PROVENANCE_KEY = "Power_Monitor_Total_Activity_Count"

RECORDING_SIZE_PER_ENTRY = 18 * BYTES_PER_WORD
DEFAULT_MALLOCS_USED = 3
CONFIG_SIZE_IN_BYTES = 2 * BYTES_PER_WORD


class ChipPowerMonitorMachineVertex(
        MachineVertex, AbstractHasAssociatedBinary,
        AbstractGeneratesDataSpecification, AbstractReceiveBuffersToHost):
    """
    Machine vertex for C code representing functionality to record
    idle times in a machine graph.

    .. note::
        This is an unusual machine vertex, in that it has no associated
        application vertex.
    """
    __slots__ = [
        "_sampling_frequency"]

    class _REGIONS(IntEnum):
        # data regions
        SYSTEM = 0
        CONFIG = 1
        RECORDING = 2

    #: which channel in the recording region has the recorded samples
    _SAMPLE_RECORDING_CHANNEL = 0

    def __init__(self, label, sampling_frequency):
        """
        :param str label: vertex label
        :param int sampling_frequency: how often to sample, in microseconds
        """
        super().__init__(
            label=label, app_vertex=None, vertex_slice=None)
        self._sampling_frequency = sampling_frequency

    @property
    def sampling_frequency(self):
        """
        How often to sample, in microseconds.

        :rtype: int
        """
        return self._sampling_frequency

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self):
        return self.get_resources(self._sampling_frequency)

    @staticmethod
    def get_resources(sampling_frequency):
        """
        Get the resources used by this vertex.

        :param float sampling_frequency:
        :rtype: ~pacman.model.resources.VariableSDRAM
        """
        # The number of sample per step does not have to be an int
        samples_per_step = (FecDataView.get_hardware_time_step_us() /
                            sampling_frequency)
        n_samples_per_recording = get_config_int(
            "EnergyMonitor", "n_samples_per_recording_entry")
        recording_per_step = (samples_per_step / n_samples_per_recording)
        max_recording_per_step = math.ceil(recording_per_step)
        overflow_recordings = max_recording_per_step - recording_per_step
        system = SYSTEM_BYTES_REQUIREMENT
        config = CONFIG_SIZE_IN_BYTES
        recording = recording_utilities.get_recording_header_size(1)
        recording += recording_utilities.get_recording_data_constant_size(1)
        fixed_sdram = system + config + recording
        with_overflow = (
            fixed_sdram + overflow_recordings * RECORDING_SIZE_PER_ENTRY)
        per_timestep = recording_per_step * RECORDING_SIZE_PER_ENTRY

        return VariableSDRAM(with_overflow, per_timestep)

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return BINARY_FILE_NAME

    @staticmethod
    def binary_file_name():
        """
        Get the filename of the binary.

        :rtype: str
        """
        return BINARY_FILE_NAME

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(
            self, spec, placement,  # @UnusedVariable
            ):
        spec.comment("\n*** Spec for ChipPowerMonitor Instance ***\n\n")

        # Construct the data images needed for the Neuron:
        self._reserve_memory_regions(spec)
        self._write_setup_info(spec)
        self._write_configuration_region(spec)

        # End-of-Spec:
        spec.end_specification()

    def _write_configuration_region(self, spec):
        """
        Write the data needed by the C code to configure itself.

        :param ~data_specification.DataSpecificationGenerator spec:
            specification writer
        """
        spec.switch_write_focus(region=self._REGIONS.CONFIG)
        n_samples_per_recording = get_config_int(
            "EnergyMonitor", "n_samples_per_recording_entry")
        spec.write_value(n_samples_per_recording, data_type=DataType.UINT32)
        spec.write_value(self._sampling_frequency, data_type=DataType.UINT32)

    def _write_setup_info(self, spec):
        """
        Writes the system data as required.

        :param ~data_specification.DataSpecificationGenerator spec:
            the DSG specification writer
        """
        spec.switch_write_focus(region=self._REGIONS.SYSTEM)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name()))

        spec.switch_write_focus(region=self._REGIONS.RECORDING)
        recorded_region_sizes = [
            self._deduce_sdram_requirements_per_timer_tick()
            * FecDataView.get_max_run_time_steps()]
        spec.write_array(recording_utilities.get_recording_header_array(
            recorded_region_sizes))

    def _reserve_memory_regions(self, spec):
        """
        Reserve the DSG memory regions as required.

        :param ~data_specification.DataSpecificationGenerator spec:
            the DSG specification to reserve in
        """
        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            region=self._REGIONS.SYSTEM,
            size=SIMULATION_N_BYTES,
            label='system')
        spec.reserve_memory_region(
            region=self._REGIONS.CONFIG,
            size=CONFIG_SIZE_IN_BYTES, label='config')
        spec.reserve_memory_region(
            region=self._REGIONS.RECORDING,
            size=recording_utilities.get_recording_header_size(1),
            label="Recording")

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return self.binary_start_type()

    @staticmethod
    def binary_start_type():
        """
        The type of binary that implements this vertex.

        :return: start-type
        :rtype: ExecutableType
        """
        return ExecutableType.USES_SIMULATION_INTERFACE

    @overrides(AbstractReceiveBuffersToHost.get_recording_region_base_address)
    def get_recording_region_base_address(self, placement):
        return locate_memory_region_for_placement(
            placement, self._REGIONS.RECORDING)

    @overrides(AbstractReceiveBuffersToHost.get_recorded_region_ids)
    def get_recorded_region_ids(self):
        return [0]

    def _deduce_sdram_requirements_per_timer_tick(self):
        """
        Deduce SDRAM usage per timer tick.

        :return: the SDRAM usage
        :rtype: int
        """
        recording_time = self._sampling_frequency * get_config_int(
            "EnergyMonitor", "n_samples_per_recording_entry")
        n_entries = math.floor(FecDataView.get_hardware_time_step_us() /
                               recording_time)
        return int(math.ceil(n_entries * RECORDING_SIZE_PER_ENTRY))

    def get_recorded_data(self, placement):
        """
        Get data from SDRAM given placement and buffer manager.
        Also arranges for provenance data to be available.

        :param ~pacman.model.placements.Placement placement:
            the location on machine to get data from
        :return: results, an array with 1 dimension of uint32 values
        :rtype: ~numpy.ndarray
        """
        # for buffering output info is taken form the buffer manager
        # get raw data as a byte array
        buffer_manager = FecDataView.get_buffer_manager()
        record_raw, data_missing = buffer_manager.get_data_by_placement(
            placement, self._SAMPLE_RECORDING_CHANNEL)
        if data_missing:
            logger.warning(
                "Chip Power monitor has lost data on chip({}, {})",
                placement.x, placement.y)

        n_samples_per_recording = get_config_int(
            "EnergyMonitor", "n_samples_per_recording_entry")
        results = (
            numpy.frombuffer(record_raw, dtype="uint32").reshape(-1, 18) /
            n_samples_per_recording)
        activity_count = int(
            numpy.frombuffer(record_raw, dtype="uint32").sum())
        with ProvenanceWriter() as db:
            db.insert_monitor(
                placement.x, placement.y, PROVENANCE_KEY, activity_count)
        return results



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import abstractproperty
from spinn_utilities.overrides import overrides
from .abstract_sends_buffers_from_host import AbstractSendsBuffersFromHost


class SendsBuffersFromHostPreBufferedImpl(
        AbstractSendsBuffersFromHost, allow_derivation=True):
    """
    Implementation of :py:class:`AbstractReceiveBuffersToHost`
    that uses an existing set of buffers for the details.
    """
    # pylint: disable=unsubscriptable-object, no-member

    __slots__ = ()

    @abstractproperty
    def send_buffers(self):
        """
        The buffer for each region that has keys to send.

        :rtype: dict(int,
            ~spinn_front_end_common.interface.buffer_management.storage_objects.BufferedSendingRegion)
        """

    @overrides(AbstractSendsBuffersFromHost.buffering_input)
    def buffering_input(self):
        return self.send_buffers is not None

    @overrides(AbstractSendsBuffersFromHost.get_regions)
    def get_regions(self):
        return self.send_buffers.keys()

    @overrides(AbstractSendsBuffersFromHost.is_next_timestamp)
    def is_next_timestamp(self, region):
        return self.send_buffers[region].is_next_timestamp

    @overrides(AbstractSendsBuffersFromHost.get_next_timestamp)
    def get_next_timestamp(self, region):
        return self.send_buffers[region].next_timestamp

    @overrides(AbstractSendsBuffersFromHost.is_next_key)
    def is_next_key(self, region, timestamp):
        return self.send_buffers[region].is_next_key(timestamp)

    @overrides(AbstractSendsBuffersFromHost.get_next_key)
    def get_next_key(self, region):
        return self.send_buffers[region].next_key

    @overrides(AbstractSendsBuffersFromHost.is_empty)
    def is_empty(self, region):
        return len(self.send_buffers[region].timestamps) == 0

    @overrides(AbstractSendsBuffersFromHost.rewind)
    def rewind(self, region):
        self.send_buffers[region].rewind()



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import struct
from collections import defaultdict
from spinn_utilities.config_holder import get_config_bool, get_config_int
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import CoreSubsets, Router
from spinnman.exceptions import (
    SpinnmanInvalidParameterException,
    SpinnmanUnexpectedResponseCodeException, SpiNNManCoresNotInStateException)
from spinnman.model import ExecutableTargets
from spinnman.model.enums import CPUState
from pacman.model.routing_tables import MulticastRoutingTables
from pacman.operations.router_compressors.ordered_covering_router_compressor\
    import (
        get_generality as
        ordered_covering_generality)
from spinn_front_end_common.abstract_models.\
    abstract_supports_bit_field_routing_compression import (
        AbstractSupportsBitFieldRoutingCompression)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.report_functions.\
    bit_field_compressor_report import (
        generate_provenance_item)
from spinn_front_end_common.utilities.exceptions import (
    CantFindSDRAMToUseException)
from spinn_front_end_common.utilities.helpful_functions import (
    get_defaultable_source_id, n_word_struct)
from spinn_front_end_common.utilities.system_control_logic import (
    run_system_application)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.constants import (
    BIT_FIELD_COMMS_SDRAM_TAG, BIT_FIELD_USABLE_SDRAM_TAG,
    BIT_FIELD_ADDRESSES_SDRAM_TAG, BIT_FIELD_ROUTING_TABLE_SDRAM_TAG)
from .load_executable_images import filter_targets
from .host_bit_field_router_compressor import (
    generate_key_to_atom_map, generate_report_path,
    start_compression_selection_process)

logger = FormatAdapter(logging.getLogger(__name__))

#: Size of SDRAM allocation for addresses
SIZE_OF_SDRAM_ADDRESS_IN_BYTES = (17 * 2 * 4) + (3 * 4)

# 7 pointers or int for each core. 4 Bytes for each  18 cores max
SIZE_OF_COMMS_SDRAM = 7 * 4 * 18

SECOND_TO_MICRO_SECOND = 1000000


class _MachineBitFieldRouterCompressor(object):
    """
    On-machine bitfield-aware routing table compression.
    """

    __slots__ = ["_compressor_aplx", "_compressor_type"]

    #: SDRAM tag the router compressor expects to find there routing tables in
    ROUTING_TABLE_SDRAM_TAG = 1

    #: SDRAM tag for the addresses the router compressor expects to find the
    #: bitfield addresses for the chip.
    BIT_FIELD_ADDRESSES_SDRAM_TAG = 2

    #
    TIMES_CYCLED_ROUTING_TABLES = 3

    #: the successful identifier
    SUCCESS = 0

    #: How many header elements are in the region addresses (1, n addresses)
    N_REGIONS_ELEMENT = 1

    #: Minimum size a heap object needs in SDRAM. (limit on the size of useful
    #: SDRAM regions to borrow)
    _MIN_SIZE_FOR_HEAP = 32

    # bit offset for compress only when needed
    _ONLY_WHEN_NEEDED_BIT_OFFSET = 1

    # bit offset for compress as much as possible
    _AS_MUCH_AS_POSS_BIT_OFFSET = 2

    # structs for performance requirements.
    _FOUR_WORDS = struct.Struct("<IIII")
    _TWO_WORDS = struct.Struct("<II")
    _ONE_WORD = struct.Struct("<I")

    # binary names
    _BIT_FIELD_SORTER_AND_SEARCH_EXECUTOR_APLX = \
        "bit_field_sorter_and_searcher.aplx"

    _HOST_BAR_TEXT = \
        "on host compressing routing tables and merging in bitfields as " \
        "appropriate"
    _ON_CHIP_ERROR_MESSAGE = \
        "The router compressor with bit field on {}:{} failed to complete. " \
        "Will execute host based routing compression instead"
    _ON_HOST_WARNING_MESSAGE = \
        "Will be executing compression for {} chips on the host, as they " \
        "failed to complete when running on chip"

    def __init__(self, compressor_aplx, compressor_type):
        self._compressor_aplx = compressor_aplx
        self._compressor_type = compressor_type

    def run(self, compress_as_much_as_possible=False):
        """
        Entrance for routing table compression with bit field.

        :param routing_tables: routing tables
        :param bool write_compressor_iobuf: flag saying if read IOBUF
        :param bool produce_report:
        :param bool compress_as_much_as_possible:
            whether to compress as much as possible
        :return: where the compressors ran
        """
        view = FecDataView()
        app_id = FecDataView.get_app_id()
        routing_tables = FecDataView.get_uncompressed()
        transceiver = FecDataView.get_transceiver()
        if len(routing_tables.routing_tables) == 0:
            return ExecutableTargets()

        # new app id for this simulation
        routing_table_compressor_app_id = view.get_new_id()

        text = f"on chip {self._compressor_type} compressor with bitfields"
        retry_count = get_config_int(
            "Mapping",
            "router_table_compression_with_bit_field_retry_count")
        if retry_count is not None:
            text += f" capped at {retry_count} retries"
        progress_bar = ProgressBar(
            total_number_of_things_to_do=(
                FecDataView.get_n_vertices() +
                (len(routing_tables.routing_tables) *
                 self.TIMES_CYCLED_ROUTING_TABLES)),
            string_describing_what_being_progressed=text)

        # locate data and on_chip_cores to load binary on
        (addresses, matrix_addresses_and_size) = self._generate_addresses(
            progress_bar)

        # create executable targets
        (compressor_executable_targets, bit_field_sorter_executable_path,
         bit_field_compressor_executable_path) = self._generate_core_subsets(
            routing_tables, progress_bar)

        # load data into sdram
        on_host_chips = self._load_data(
            addresses, transceiver, routing_table_compressor_app_id,
            routing_tables, app_id,
            compress_as_much_as_possible, progress_bar,
            compressor_executable_targets,
            matrix_addresses_and_size, bit_field_compressor_executable_path,
            bit_field_sorter_executable_path, retry_count)

        # load and run binaries
        try:
            run_system_application(
                compressor_executable_targets,
                routing_table_compressor_app_id,
                get_config_bool("Reports", "write_compressor_iobuf"),
                functools.partial(
                    self._check_bit_field_router_compressor_for_success,
                    host_chips=on_host_chips,
                    sorter_binary_path=bit_field_sorter_executable_path),
                [CPUState.FINISHED], True,
                "bit_field_compressor_on_{}_{}_{}.txt",
                [bit_field_sorter_executable_path], progress_bar,
                logger=logger)
        except SpiNNManCoresNotInStateException as e:
            logger.exception(transceiver.get_core_status_string(
                e.failed_core_states()))
            try:
                transceiver.stop_application(routing_table_compressor_app_id)
            except Exception:  # pylint: disable=broad-except
                logger.warning("Could not stop compressor!")
            raise e

        # start the host side compressions if needed
        if len(on_host_chips) != 0:
            most_costly_cores = defaultdict(lambda: defaultdict(int))
            for partition in FecDataView.iterate_partitions():
                for edge in partition.edges:
                    sttr = edge.pre_vertex.splitter
                    for vertex in sttr.get_source_specific_in_coming_vertices(
                            partition.pre_vertex, partition.identifier):
                        place = FecDataView.get_placement_of_vertex(vertex)
                        if place.chip in on_host_chips:
                            most_costly_cores[place.chip][place.p] += 1
            logger.warning(self._ON_HOST_WARNING_MESSAGE, len(on_host_chips))
            progress_bar = ProgressBar(
                total_number_of_things_to_do=len(on_host_chips),
                string_describing_what_being_progressed=self._HOST_BAR_TEXT)
            compressed_pacman_router_tables = MulticastRoutingTables()

            key_atom_map = generate_key_to_atom_map()

            for (chip_x, chip_y) in progress_bar.over(on_host_chips, False):
                if get_config_bool(
                        "Reports",
                        "write_router_compression_with_bitfield_report"):
                    report_folder_path = generate_report_path()
                else:
                    report_folder_path = None

                start_compression_selection_process(
                    router_table=routing_tables.get_routing_table_for_chip(
                        chip_x, chip_y),
                    report_folder_path=report_folder_path,
                    most_costly_cores=most_costly_cores,
                    compressed_pacman_router_tables=(
                        compressed_pacman_router_tables),
                    key_atom_map=key_atom_map)

            # load host compressed routing tables
            for table in compressed_pacman_router_tables.routing_tables:
                if table.multicast_routing_entries:
                    transceiver.clear_multicast_routes(table.x, table.y)
                    transceiver.load_multicast_routes(
                        table.x, table.y, table.multicast_routing_entries,
                        app_id=app_id)

            progress_bar.end()

        return compressor_executable_targets

    def _generate_core_subsets(self, routing_tables, progress_bar):
        """
        Generates the core subsets for the binaries.

        :param ~.MulticastRoutingTables routing_tables: the routing tables
        :param ~.ProgressBar progress_bar: progress bar
        :param ~spinnman.model.ExecutableTargets system_executable_targets:
            the executables targets to cores
        :return: (targets, sorter path, and compressor path)
        :rtype: tuple(~spinnman.model.ExecutableTargets, str, str)
        """
        bit_field_sorter_cores = CoreSubsets()
        bit_field_compressor_cores = CoreSubsets()

        _, cores = filter_targets(lambda ty: ty is ExecutableType.SYSTEM)
        view = FecDataView()
        for routing_table in progress_bar.over(routing_tables, False):
            # add 1 core to the sorter, and the rest to compressors
            sorter = None
            for processor in view.get_chip_at(
                    routing_table.x, routing_table.y).processors:
                if (not processor.is_monitor and
                        not cores.all_core_subsets.is_core(
                            routing_table.x, routing_table.y,
                            processor.processor_id)):
                    if sorter is None:
                        sorter = processor
                        bit_field_sorter_cores.add_processor(
                            routing_table.x, routing_table.y,
                            processor.processor_id)
                    else:
                        bit_field_compressor_cores.add_processor(
                            routing_table.x, routing_table.y,
                            processor.processor_id)

        # convert core subsets into executable targets
        executable_targets = ExecutableTargets()

        # bit field executable paths
        bit_field_sorter_executable_path = \
            FecDataView.get_executable_path(
                self._BIT_FIELD_SORTER_AND_SEARCH_EXECUTOR_APLX)

        bit_field_compressor_executable_path = \
            FecDataView.get_executable_path(self._compressor_aplx)

        # add the sets
        executable_targets.add_subsets(
            binary=bit_field_sorter_executable_path,
            subsets=bit_field_sorter_cores,
            executable_type=ExecutableType.SYSTEM)
        executable_targets.add_subsets(
            binary=bit_field_compressor_executable_path,
            subsets=bit_field_compressor_cores,
            executable_type=ExecutableType.SYSTEM)

        return (executable_targets, bit_field_sorter_executable_path,
                bit_field_compressor_executable_path)

    def _check_bit_field_router_compressor_for_success(
            self, executable_targets, host_chips, sorter_binary_path):
        """
        Goes through the cores checking for cores that have failed to
        generate the compressed routing tables with bitfield.

        :param ~spinnman.model.ExecutableTargets executable_targets:
            cores to load router compressor with bitfield on
        :param list(tuple(int,int)) host_chips:
            the chips which need to be ran on host.
        :param str sorter_binary_path: the path to the sorter binary
        :rtype: bool
        """
        transceiver = FecDataView.get_transceiver()
        sorter_cores = executable_targets.get_cores_for_binary(
            sorter_binary_path)
        result = True
        for core_subset in sorter_cores:
            x = core_subset.x
            y = core_subset.y

            for p in core_subset.processor_ids:
                # Read the result from USER1/USER2 registers
                user_1_base_address = \
                    transceiver.get_user_1_register_address_from_core(p)
                user_2_base_address = \
                    transceiver.get_user_2_register_address_from_core(p)
                result = transceiver.read_word(x, y, user_1_base_address)
                bit_fields_merged = transceiver.read_word(
                    x, y, user_2_base_address)

                if result != self.SUCCESS:
                    host_chips.add((x, y))
                    result = False
                generate_provenance_item(x, y, bit_fields_merged)
        return result

    def _load_data(
            self, addresses, transceiver, routing_table_compressor_app_id,
            routing_tables, app_id,
            compress_as_much_as_possible, progress_bar, cores,
            matrix_addresses_and_size,
            bit_field_compressor_executable_path,
            bit_field_sorter_executable_path, retry_count):
        """
        load all data onto the chip.

        :param dict(tuple(int,int),tuple(int,int)) addresses:
            the addresses for bitfields in SDRAM
        :param ~.Transceiver transceiver: the spinnMan instance
        :param routing_table_compressor_app_id:
            the app_id for the system application
        :param ~.MulticastRoutingTables routing_tables:
            the routing tables
        :param int app_id: the app_id of the application
        :param ~.ProgressBar progress_bar: progress bar
        :param bool compress_as_much_as_possible:
            whether to compress as much as possible
        :param ~spinnman.model.ExecutableTargets cores:
            the cores that compressor will run on
        :param dict matrix_addresses_and_size:
            maps chips to regeneration SDRAM and size for exploitation
        :param str bit_field_compressor_executable_path:
            the path to the compressor binary path
        :param str bit_field_sorter_executable_path:
            the path to the sorter binary
        :param retry_count:
            Number of times that the sorters should set of the compressions
            again. `None` for as much as needed
        :type retry_count: int or None
        :return:
            the list of tuples saying which chips this will need to use
            host compression, as the malloc failed.
        :rtype: list(tuple(int,int))
        """
        run_by_host = list()
        for table in routing_tables.routing_tables:
            try:
                self._load_routing_table_data(
                    table, app_id, transceiver,
                    routing_table_compressor_app_id, progress_bar, cores,
                    matrix_addresses_and_size[(table.x, table.y)])

                comms_sdram = transceiver.malloc_sdram(
                    table.x, table.y, SIZE_OF_COMMS_SDRAM,
                    routing_table_compressor_app_id,
                    BIT_FIELD_COMMS_SDRAM_TAG)

                self._load_address_data(
                    addresses, table.x, table.y, transceiver,
                    routing_table_compressor_app_id,
                    cores, matrix_addresses_and_size[(table.x, table.y)],
                    bit_field_compressor_executable_path,
                    bit_field_sorter_executable_path, comms_sdram,
                    retry_count)

                self._load_usable_sdram(
                    matrix_addresses_and_size[(table.x, table.y)], table.x,
                    table.y, transceiver, routing_table_compressor_app_id,
                    cores)

                self._load_compressor_data(
                    table.x, table.y, transceiver,
                    bit_field_compressor_executable_path, cores,
                    compress_as_much_as_possible, comms_sdram)
            except CantFindSDRAMToUseException:
                run_by_host.append((table.x, table.y))

        return run_by_host

    def _load_compressor_data(
            self, chip_x, chip_y, transceiver,
            bit_field_compressor_executable_path, cores,
            compress_as_much_as_possible, comms_sdram):
        """
        Updates the user addresses for the compressor cores with the
        compression settings.

        :param int chip_x: chip X coordinate
        :param int chip_y: chip Y coordinate
        :param ~spinnman.transceiver.Transceiver transceiver: SpiNNMan instance
        :param str bit_field_compressor_executable_path:
            path for the compressor binary
        :param bool compress_as_much_as_possible:
            whether to compress as much as possible
        :param ~spinnman.model.ExecutableTargets cores: the executable targets
        :param int comms_sdram: Address for communications block
        """
        compressor_cores = cores.get_cores_for_binary(
            bit_field_compressor_executable_path)
        for processor_id in compressor_cores.get_core_subset_for_chip(
                chip_x, chip_y).processor_ids:
            user1_address = \
                transceiver.get_user_1_register_address_from_core(processor_id)
            user2_address = \
                transceiver.get_user_2_register_address_from_core(processor_id)
            user3_address = \
                transceiver.get_user_3_register_address_from_core(processor_id)
            # user 1 the time per compression attempt
            time_per_iteration = get_config_int(
                "Mapping",
                "router_table_compression_with_bit_field_iteration_time")
            transceiver.write_memory(
                chip_x, chip_y, user1_address,
                int(time_per_iteration * SECOND_TO_MICRO_SECOND))
            # user 2 Compress as much as needed flag
            transceiver.write_memory(
                chip_x, chip_y, user2_address,
                int(compress_as_much_as_possible))
            # user 3 the comms_sdram area
            transceiver.write_memory(
                chip_x, chip_y, user3_address, comms_sdram)

    def _load_usable_sdram(
            self, matrix_addresses_and_size, chip_x, chip_y, transceiver,
            routing_table_compressor_app_id, cores):
        """
        loads the addresses of borrowable SDRAM.

        :param list(tuple(int,int)) matrix_addresses_and_size:
            SDRAM usable and sizes
        :param int chip_x: X coordinate of the chip to consider here
        :param int chip_y: Y coordinate of the chip to consider here
        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnman instance
        :param int routing_table_compressor_app_id: system app_id.
        :param ~spinnman.model.ExecutableTargets cores:
            the cores that compressor will run on
        """
        address_data = self._generate_chip_matrix_data(
            matrix_addresses_and_size)

        # get sdram address on chip
        try:
            sdram_address = transceiver.malloc_sdram(
                chip_x, chip_y, len(address_data),
                routing_table_compressor_app_id,
                BIT_FIELD_USABLE_SDRAM_TAG)
        except (SpinnmanInvalidParameterException,
                SpinnmanUnexpectedResponseCodeException):
            sdram_address = self._borrow_from_matrix_addresses(
                matrix_addresses_and_size, len(address_data))
            address_data = self._generate_chip_matrix_data(
                matrix_addresses_and_size)

        # write sdram
        transceiver.write_memory(
            chip_x, chip_y, sdram_address, address_data, len(address_data))

        # get the only processor on the chip
        processor_id = list(cores.all_core_subsets.get_core_subset_for_chip(
            chip_x, chip_y).processor_ids)[0]

        # update user 2 with location
        user3_address = transceiver.get_user_3_register_address_from_core(
            processor_id)
        transceiver.write_memory(
            chip_x, chip_y, user3_address, sdram_address)

    def _generate_chip_matrix_data(self, list_of_sizes_and_address):
        """
        Generate the data for the chip matrix data.

        :param list(tuple(int,int)) list_of_sizes_and_address:
            SDRAM addresses and sizes
        :return: byte array of data
        """
        data = b""
        data += self._ONE_WORD.pack(len(list_of_sizes_and_address))
        for (memory_address, size) in list_of_sizes_and_address:
            data += self._TWO_WORDS.pack(memory_address, size)
        return data

    def _load_address_data(
            self, addresses, chip_x, chip_y, transceiver,
            routing_table_compressor_app_id, cores, matrix_addresses_and_size,
            bit_field_compressor_executable_path,
            bit_field_sorter_executable_path, comms_sdram, retry_count):
        """
        loads the bitfield addresses space.

        :param dict(tuple(int,int),tuple(int,int)) addresses:
            the addresses to load
        :param int chip_x: the chip x to consider here
        :param int chip_y: the chip y to consider here
        :param ~spinnman.transceiver.Transceiver transceiver:
            the transceiver instance
        :param int routing_table_compressor_app_id: system app_id.
        :param ~spinnman.model.ExecutableTargets cores:
            the cores that compressor will run on
        :param str bit_field_compressor_executable_path:
            the path to the compressor binary path
        :param str bit_field_sorter_executable_path:
            the path to the sorter binary
        :param int comms_sdram: Address for communications block
        :param retry_count:
            Number of times that the sorters should set of the compressions
            again. `None` for as much as needed
        :type retry_count: int or None
        """
        # generate address_data
        address_data = self._generate_chip_data(
            addresses[(chip_x, chip_y)],
            cores.get_cores_for_binary(
                bit_field_compressor_executable_path).get_core_subset_for_chip(
                    chip_x, chip_y),
            comms_sdram, retry_count)

        # get sdram address on chip
        try:
            sdram_address = transceiver.malloc_sdram(
                chip_x, chip_y, len(address_data),
                routing_table_compressor_app_id,
                BIT_FIELD_ADDRESSES_SDRAM_TAG)
        except (SpinnmanInvalidParameterException,
                SpinnmanUnexpectedResponseCodeException):
            sdram_address = self._borrow_from_matrix_addresses(
                matrix_addresses_and_size, len(address_data))

        # write sdram
        transceiver.write_memory(
            chip_x, chip_y, sdram_address, address_data, len(address_data))

        # get the only processor on the chip
        sorter_cores = cores.get_cores_for_binary(
            bit_field_sorter_executable_path)
        processor_id = list(sorter_cores.get_core_subset_for_chip(
            chip_x, chip_y).processor_ids)[0]

        # update user 2 with location
        user2_address = transceiver.get_user_2_register_address_from_core(
            processor_id)
        transceiver.write_memory(
            chip_x, chip_y, user2_address, sdram_address)

    def _load_routing_table_data(
            self, table, app_id, transceiver,
            routing_table_compressor_app_id, progress_bar, cores,
            matrix_addresses_and_size):
        """
        loads the routing table data.

        :param table: the routing table to load
        :type table:
            ~pacman.model.routing_tables.AbstractMulticastRoutingTable
        :param int app_id: application app_id
        :param ~spinnman.transceiver.Transceiver transceiver:
            transceiver instance
        :param ~spinn_utilities.progress_bar.ProgressBar progress_bar:
            progress bar
        :param int routing_table_compressor_app_id: system app_id
        :param ~spinnman.model.ExecutableTargets cores:
            the cores that the compressor going to run on
        :raises CantFindSDRAMToUse: when SDRAM is not malloc'ed or stolen
        """
        routing_table_data = self._build_routing_table_data(app_id, table)

        # go to spinnman and ask for a memory region of that size per chip.
        try:
            base_address = transceiver.malloc_sdram(
                table.x, table.y, len(routing_table_data),
                routing_table_compressor_app_id,
                BIT_FIELD_ROUTING_TABLE_SDRAM_TAG)
        except (SpinnmanInvalidParameterException,
                SpinnmanUnexpectedResponseCodeException):
            base_address = self._borrow_from_matrix_addresses(
                matrix_addresses_and_size, len(routing_table_data))

        # write SDRAM requirements per chip
        transceiver.write_memory(
            table.x, table.y, base_address, routing_table_data)

        # get the only processor on the chip
        processor_id = list(cores.all_core_subsets.get_core_subset_for_chip(
            table.x, table.y).processor_ids)[0]

        # update user 1 with location
        user1_address = transceiver.get_user_1_register_address_from_core(
            processor_id)
        transceiver.write_memory(
            table.x, table.y, user1_address, base_address)

        # update progress bar
        progress_bar.update()

    def _build_routing_table_data(self, app_id, routing_table):
        """
        Builds routing data as needed for the compressor cores.

        :param int app_id: app_id of the application to load entries with
        :param ~.AbstractMulticastRoutingTable routing_table:
            the uncompressed routing table
        :return: data array
        :rtype: bytearray
        """
        data = b''
        data += self._TWO_WORDS.pack(app_id, routing_table.number_of_entries)

        # sort entries based on generality
        sorted_routing_table = sorted(
            routing_table.multicast_routing_entries,
            key=lambda rt_entry: ordered_covering_generality(
                rt_entry.routing_entry_key, rt_entry.mask))

        # write byte array for the sorted table
        for entry in sorted_routing_table:
            data += self._FOUR_WORDS.pack(
                entry.routing_entry_key, entry.mask,
                Router.convert_routing_table_entry_to_spinnaker_route(entry),
                get_defaultable_source_id(entry=entry))
        return bytearray(data)

    @staticmethod
    def _borrow_from_matrix_addresses(
            matrix_addresses_and_size, size_to_borrow):
        """
        Borrows memory from synaptic matrix as needed.

        :param dict(tuple(int,int),tuple(int,int)) matrix_addresses_and_size:
            matrix addresses and sizes; updated by this method
        :param int size_to_borrow: size needed to borrow from matrices.
        :return: address to start borrow from
        :rtype: int
        :raises CantFindSDRAMToUseException:
            when no space is big enough to borrow from.
        """
        for pos, (base_address, size) in enumerate(matrix_addresses_and_size):
            if size >= size_to_borrow:
                new_size = size - size_to_borrow
                matrix_addresses_and_size[pos] = (base_address, new_size)
                return base_address
        raise CantFindSDRAMToUseException()

    def _add_to_addresses(
            self, vertex, placement, region_addresses,
            sdram_block_addresses_and_sizes):
        """
        Adds data about the API-based vertex.

        :param AbstractSupportsBitFieldRoutingCompression vertex:
            vertex which utilises the API
        :param ~.Placement placement: placement of vertex
        :param dict(tuple(int,int),list(tuple(int,int))) region_addresses:
            store for data regions
        :param sdram_block_addresses_and_sizes: store for surplus SDRAM
        :type sdram_block_addresses_and_sizes:
            dict(tuple(int,int),list(tuple(int,int)))
        """
        # store the region sdram address's
        bit_field_sdram_address = vertex.bit_field_base_address(placement)
        region_addresses[placement.x, placement.y].append(
            (bit_field_sdram_address, placement.p))

        # store the available space from the matrix to borrow
        blocks = vertex.regeneratable_sdram_blocks_and_sizes(placement)

        for (address, size) in blocks:
            if size != 0 and size > self._MIN_SIZE_FOR_HEAP:
                sdram_block_addresses_and_sizes[
                    placement.x, placement.y].append((address, size))
        sorted(
            sdram_block_addresses_and_sizes[placement.x, placement.y],
            key=lambda data: data[0])

    def _generate_addresses(self, progress_bar):
        """
        Generates the bitfield SDRAM addresses.

        :param ~.ProgressBar progress_bar: the progress bar
        :return: region_addresses and the executable targets to load the
            router table compressor with bitfield. and the SDRAM blocks
            available for use on each core that we plan to use
        :rtype: tuple(dict(tuple(int,int),tuple(int,int)),
            dict(tuple(int,int),list(tuple(int,int))))
        """
        # data holders
        region_addresses = defaultdict(list)
        sdram_block_addresses_and_sizes = defaultdict(list)

        for app_vertex in progress_bar.over(
                FecDataView.iterate_vertices(), finish_at_end=False):
            for m_vertex in app_vertex.machine_vertices:
                if isinstance(
                        m_vertex, AbstractSupportsBitFieldRoutingCompression):
                    placement = FecDataView.get_placement_of_vertex(m_vertex)
                    self._add_to_addresses(
                        m_vertex, placement, region_addresses,
                        sdram_block_addresses_and_sizes)
        return region_addresses, sdram_block_addresses_and_sizes

    def _generate_chip_data(
            self, address_list, cores, comms_sdram, retry_count):
        """
        Generate the region_addresses_t data.

        * Minimum percentage of bitfields to be merge in (currently ignored)

        * Number of times that the sorters should set of the compressions again

        * Pointer to the area malloc'ed to hold the comms_sdram

        * Number of processors in the list

        * The data for the processors

        :param list(tuple(int,int)) address_list:
            the list of SDRAM addresses
        :param ~.CoreSubset cores: compressor cores on this chip.
        :param int comms_sdram: Address for communications block
        :param retry_count:
            Number of times that the sorters should set of the compressions
            again. `None` for as much as needed
        :type retry_count: int or None
        :return: the byte array
        :rtype: bytes
        """
        threshold_percentage = get_config_int(
            "Mapping",
            "router_table_compression_with_bit_field_acceptance_threshold")

        data = self._FOUR_WORDS.pack(
            threshold_percentage,
            retry_count if retry_count is not None else 0xFFFFFFFF,
            comms_sdram, len(address_list))
        for (bit_field, processor_id) in address_list:
            data += self._TWO_WORDS.pack(bit_field, processor_id)
        data += self._ONE_WORD.pack(len(cores))
        data += n_word_struct(len(cores)).pack(*list(cores.processor_ids))
        return data


def machine_bit_field_ordered_covering_compressor(
        compress_as_much_as_possible=False):
    """
    Compression with bit field and ordered covering.

    :param bool compress_as_much_as_possible:
        whether to compress as much as possible
    :return: where the compressors ran
    """
    compressor = _MachineBitFieldRouterCompressor(
        "bit_field_ordered_covering_compressor.aplx", "OrderedCovering")
    return compressor.run(compress_as_much_as_possible)


def machine_bit_field_pair_router_compressor(
        compress_as_much_as_possible=False):
    """
    Compression with bit field pairing.

    :param bool compress_as_much_as_possible:
        whether to compress as much as possible
    :return: where the compressors ran
    """
    compressor = _MachineBitFieldRouterCompressor(
        "bit_field_pair_compressor.aplx", "Pair")
    return compressor.run(compress_as_much_as_possible)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
import logging
import numpy
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinn_machine import CoreSubsets
from data_specification import DataSpecificationExecutor, MemoryRegionReal
from data_specification.constants import (
    MAX_MEM_REGIONS, APP_PTR_TABLE_BYTE_SIZE)
from data_specification.exceptions import DataSpecificationException
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.helpful_functions import (
    write_address_to_user0)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.emergency_recovery import (
    emergency_recover_states_from_failure)
from spinn_front_end_common.utilities.constants import CORE_DATA_SDRAM_BASE_TAG

logger = FormatAdapter(logging.getLogger(__name__))
_MEM_REGIONS = range(MAX_MEM_REGIONS)


def system_cores():
    """
    Get the subset of cores that are to be used for system operations.

    :rtype: ~spinn_machine.CoreSubsets
    """
    cores = CoreSubsets()
    exec_targets = FecDataView.get_executable_targets()
    for binary in exec_targets.get_binaries_of_executable_type(
            ExecutableType.SYSTEM):
        cores.add_core_subsets(exec_targets.get_cores_for_binary(binary))
    return cores


#: A named tuple for a region that can be referenced
_RegionToRef = namedtuple(
    "__RegionToUse", ["x", "y", "p", "region", "pointer"])


#: A class for regions to be filled in
class _CoreToFill(object):
    __slots__ = [
        "x", "y", "p", "header", "pointer_table", "base_address", "regions"]

    def __init__(self, x, y, p, header, pointer_table, base_address):
        self.x = x
        self.y = y
        self.p = p
        self.header = header
        self.pointer_table = pointer_table
        self.base_address = base_address
        self.regions = []


class _ExecutionContext(object):
    """
    A context for executing multiple data specifications with
    cross-references.
    """

    def __init__(self):
        self.__references_to_fill = list()
        self.__references_to_use = dict()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Only do the close bit if nothing has gone wrong
        if exc_type is None:
            self.close()
        # Force any exception to be raised
        return False

    def execute(
            self, core, reader, writer_func, base_address, size_allocated):
        """
        Execute the data spec for a core.

        :param tuple(int,int,int) core:
        :param ~io.RawIOBase reader:
        :param int base_address:
        :param int size_allocated:
        :return: base_address, size_allocated, bytes_written
        :rtype: tuple(int, int, int)
        """
        x, y, p = core

        # Maximum available memory.
        # However, system updates the memory available independently, so the
        # space available check actually happens when memory is allocated.
        memory_available = FecDataView().get_chip_at(x, y).sdram.size

        # generate data spec executor
        executor = DataSpecificationExecutor(reader, memory_available)

        # run data spec executor
        try:
            executor.execute()
        except DataSpecificationException:
            logger.error("Error executing data specification for {}, {}, {}",
                         x, y, p)
            raise

        # Write the header and pointer table
        header = executor.get_header()
        pointer_table = executor.get_pointer_table(base_address)

        # Handle references to the regions from the executor
        self.__handle_new_references(x, y, p, executor, pointer_table)

        # Handle references that need to be filled
        write_header_now = self.__handle_references_to_fill(
            x, y, p, executor, pointer_table, header, base_address)

        # We don't have to write this bit now; we can still to the rest
        if write_header_now:
            # NB: DSE meta-block is always small (i.e., one SDP write)

            to_write = numpy.concatenate(
                (header, pointer_table.view("uint32"))).tobytes()
            FecDataView.write_memory(x, y, base_address, to_write)

        # Write each region
        bytes_written = APP_PTR_TABLE_BYTE_SIZE
        for region_id in _MEM_REGIONS:
            region = executor.get_region(region_id)
            if not isinstance(region, MemoryRegionReal):
                continue
            max_pointer = region.max_write_pointer
            if region.unfilled or max_pointer == 0:
                continue

            # Get the data up to what has been written
            data = region.region_data[:max_pointer]

            # Write the data to the position
            writer_func(x, y, pointer_table[region_id]["pointer"], data)
            bytes_written += len(data)

        return base_address, size_allocated, bytes_written

    def close(self):
        """
        Called when finished executing all regions.  Fills in the
        references if possible, and fails if not.
        """
        for core_to_fill in self.__references_to_fill:
            pointer_table = core_to_fill.pointer_table
            for ref_region, ref in core_to_fill.regions:
                if ref not in self.__references_to_use:
                    raise ValueError(
                        f"Reference {ref} requested from {core_to_fill} "
                        "but not found")
                pointer_table[ref_region]["pointer"] = self.__get_reference(
                    ref, core_to_fill.x, core_to_fill.y, core_to_fill.p,
                    ref_region)
            to_write = numpy.concatenate(
                (core_to_fill.header, pointer_table.view("uint32"))).tobytes()
            FecDataView.write_memory(
                core_to_fill.x, core_to_fill.y, core_to_fill.base_address,
                to_write)

    def __handle_new_references(self, x, y, p, executor, pointer_table):
        """
        Get references that can be used later.

        :param int x: The x-coordinate of the spec being executed
        :param int y: The y-coordinate of the spec being executed
        :param int p: The core of the spec being executed
        :param ~DataSpecificationExecutor executor:
            The execution context
        :param list(int) pointer_table:
            A table of pointers that could be referenced later
        """
        for ref_region in executor.referenceable_regions:
            ref = executor.get_region(ref_region).reference
            if ref in self.__references_to_use:
                ref_to_use = self.__references_to_use[ref]
                raise ValueError(
                    f"Reference {ref} used previously as {ref_to_use} so "
                    f"cannot be used by {x}, {y}, {p}, {ref_region}")
            ptr = pointer_table[ref_region]["pointer"]
            self.__references_to_use[ref] = _RegionToRef(
                x, y, p, ref_region, ptr)

    def __handle_references_to_fill(
            self, x, y, p, executor, pointer_table, header, base_address):
        """
        Resolve references.

        :param int x: The x-coordinate of the spec being executed
        :param int y: The y-coordinate of the spec being executed
        :param int p: The core of the spec being executed
        :param ~DataSpecificationExecutor executor:
            The execution context
        :param list(int) pointer_table:
            A table of pointers that could be referenced later
        :param header: The Data Specification header bytes
        :param base_address: The base address of the executed spec
        :return: Whether all references were resolved
        """
        # Resolve any references now
        coreToFill = _CoreToFill(x, y, p, header, pointer_table, base_address)
        for ref_region in executor.references_to_fill:
            ref = executor.get_region(ref_region).ref
            # If already been created, use directly
            if ref in self.__references_to_use:
                pointer_table[ref_region]["pointer"] = self.__get_reference(
                    ref, x, y, p, ref_region)
            else:
                coreToFill.regions.append((ref_region, ref))
        if coreToFill.regions:
            self.__references_to_fill.append(coreToFill)
        return not bool(coreToFill.regions)

    def __get_reference(self, ref, x, y, p, ref_region):
        """
        Get a reference to a region, doing some extra checks on eligibility.

        :param int ref: The reference to the region
        :param int x: The x-coordinate of the executing spec
        :param int y: The y-coordinate of the executing spec
        :param int p: The core of the executing spec
        :param .CoreToFill ref_region: Data related to the reference
        :return: The pointer to use
        :raise ValueError:
            if the reference cannot be referenced in this context
        """
        ref_to_use = self.__references_to_use[ref]
        if ref_to_use.x != x or ref_to_use.y != y:
            raise ValueError(
                f"Reference {ref} to {ref_to_use} cannot be used by "
                f"{x}, {y}, {p}, {ref_region} because they are on "
                "different chips")
        return ref_to_use.pointer


def execute_system_data_specs():
    """
    Execute the data specs for all system targets.
    """
    specifier = _HostExecuteDataSpecification()
    return specifier.execute_system_data_specs()


def execute_application_data_specs():
    """
    Execute the data specs for all non-system targets.
    """
    specifier = _HostExecuteDataSpecification()
    specifier.execute_application_data_specs()


class _HostExecuteDataSpecification(object):
    """
    Executes the host based data specification.
    """

    __slots__ = [
        # the application ID of the simulation
        "_app_id"]

    first = True

    def __init__(self):
        self._app_id = FecDataView.get_app_id()

    def execute_application_data_specs(self):
        """
        Execute the data specs for all non-system targets.
        """

        uses_advanced_monitors = get_config_bool(
            "Machine", "enable_advanced_monitor_support")
        # Allow config to override
        if get_config_bool(
                "Machine", "disable_advanced_monitor_usage_for_data_in"):
            uses_advanced_monitors = False
        try:
            if FecDataView.has_java_caller():
                return self.__java_app(uses_advanced_monitors)
            else:
                return self.__python_app(uses_advanced_monitors)
        except:  # noqa: E722
            if uses_advanced_monitors:
                emergency_recover_states_from_failure()
            raise

    def __set_router_timeouts(self):
        for receiver in FecDataView.iterate_gathers():
            receiver.load_system_routing_tables()
            receiver.set_cores_for_data_streaming()

    def __reset_router_timeouts(self):
        # reset router timeouts
        for receiver in FecDataView.iterate_gathers():
            receiver.unset_cores_for_data_streaming()
            # reset router tables
            receiver.load_application_routing_tables()

    def __select_writer(self, x, y):
        view = FecDataView()
        chip = view.get_chip_at(x, y)
        ethernet_chip = view.get_chip_at(
            chip.nearest_ethernet_x, chip.nearest_ethernet_y)
        gatherer = view.get_gatherer_by_xy(ethernet_chip.x, ethernet_chip.y)
        return gatherer.send_data_into_spinnaker

    def __python_app(self, use_monitors):
        """
        :param bool use_monitors:
        """
        if use_monitors:
            self.__set_router_timeouts()

        dsg_targets = FecDataView.get_dsg_targets()
        # create a progress bar for end users
        progress = ProgressBar(
            dsg_targets.ds_n_app_cores(),
            "Executing data specifications and loading data for "
            "application vertices")

        with _ExecutionContext() as context:
            transceiver = FecDataView.get_transceiver()
            for core, reader, region_size in progress.over(
                    dsg_targets.app_items()):
                x, y, p = core
                base_address = self.__malloc_region_storage(
                    core, region_size)
                base_address, size_allocated, bytes_written = context.execute(
                    core, reader, self.__select_writer(x, y)
                    if use_monitors else transceiver.write_memory,
                    base_address, region_size)
                dsg_targets.set_write_info(
                    x, y, p, base_address, size_allocated, bytes_written)

        if use_monitors:
            self.__reset_router_timeouts()

    def __java_app(self, use_monitors):
        """
        :param bool use_monitors:
        """
        # create a progress bar for end users
        progress = ProgressBar(
            2, "Executing data specifications and loading data for "
            "application vertices using Java")

        java_caller = FecDataView.get_java_caller()
        if use_monitors:
            # Method also called with just recording params
            java_caller.set_placements(FecDataView.iterate_placemements())
        progress.update()

        java_caller.execute_app_data_specification(use_monitors)
        progress.end()

    def execute_system_data_specs(self):
        """
        Execute the data specs for all system targets.
        """
        # pylint: disable=too-many-arguments
        FecDataView.get_dsg_targets().mark_system_cores(system_cores())
        if FecDataView.has_java_caller():
            self.__java_sys()
        else:
            self.__python_sys()

    def __java_sys(self):
        """
        Does the Data Specification Execution and loading using Java.
        """
        # create a progress bar for end users
        progress = ProgressBar(
            1, "Executing data specifications and loading data for system "
            "vertices using Java")
        FecDataView.get_java_caller().execute_system_data_specification()
        progress.end()

    def __python_sys(self):
        """
        Does the Data Specification Execution and loading using Python.
        """

        # create a progress bar for end users
        dsg_targets = FecDataView.get_dsg_targets()
        progress = ProgressBar(
            dsg_targets.ds_n_system_cores(),
            "Executing data specifications and loading data for "
            "system vertices")

        # allocate and set user 0 before loading data

        with _ExecutionContext() as context:
            transceiver = FecDataView.get_transceiver()
            for core, reader, region_size in progress.over(
                    dsg_targets.system_items()):
                x, y, p = core
                base_address = self.__malloc_region_storage(
                    core, region_size)
                base_address, size_allocated, bytes_written = context.execute(
                    core, reader, transceiver.write_memory,
                    base_address, region_size)
                dsg_targets.set_write_info(
                    x, y, p, base_address, size_allocated, bytes_written)

    def __malloc_region_storage(self, core, size):
        """
        Allocates the storage for all DSG regions on the core and tells
        the core and our caller where that storage is.

        :param tuple(int,int,int) core: Which core we're talking about.
        :param int size:
            The total size of all storage for regions on that core, including
            for the header metadata.
        :return: address of region header table (not yet filled)
        :rtype: int
        """
        (x, y, p) = core

        # allocate memory where the app data is going to be written; this
        # raises an exception in case there is not enough SDRAM to allocate
        start_address = FecDataView.get_transceiver().malloc_sdram(
            x, y, size, self._app_id, tag=CORE_DATA_SDRAM_BASE_TAG + p)

        # set user 0 register appropriately to the application data
        write_address_to_user0(x, y, p, start_address)

        return start_address



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from .dpri_flags import DPRIFlags

_PATTERN = struct.Struct("<IIIIIIIIII")


def _decode_router_timeout_value(value):
    """
    Get the timeout value of a router in ticks, given an 8-bit floating
    point value stored in an int (!)

    :param int value: The value to convert
    :rtype: int
    """
    mantissa = value & 0xF
    exponent = (value >> 4) & 0xF
    if exponent <= 4:
        return ((mantissa + 16) - (2 ** (4 - exponent))) * (2 ** exponent)
    return (mantissa + 16) * (2 ** exponent)


class ReInjectionStatus(object):
    """
    Represents a status information report from dropped packet reinjection.
    """

    __slots__ = (

        # The WAIT1 timeout value of the router in cycles
        "_wait1_timeout",

        # The WAIT2 timeout value of the router in cycles
        "_wait2_timeout",

        # The number of packets dropped by the router and received by\
        # the re injection functionality (may not fit in the queue though)
        "_n_dropped_packets",

        # The number of times that when a dropped packet was read it was\
        #    found that another one or more packets had also been dropped,\
        #    but had been missed
        "_n_missed_dropped_packets",

        # Of the n_dropped_packets received, how many were lost due to not\
        #    having enough space in the queue of packets to reinject
        "_n_dropped_packet_overflows",

        # Of the n_dropped_packets received, how many packets were\
        #    successfully re-injected
        "_n_reinjected_packets",

        # The number of times that when a dropped packet was caused due to\
        # a link failing to take the packet.
        "_n_link_dumps",

        # The number of times that when a dropped packet was caused due to\
        # a processor failing to take the packet.
        "_n_processor_dumps",

        # the flags that states which types of packets were being recorded
        "_flags",

        # Indicates the links or processors dropped from
        "_link_proc_bits"
    )

    def __init__(self, data, offset):
        """
        :param bytes data: The data containing the information
        :param int offset: The offset in the data where the information starts
        """
        (self._wait1_timeout, self._wait2_timeout,
         self._n_dropped_packets, self._n_missed_dropped_packets,
         self._n_dropped_packet_overflows, self._n_reinjected_packets,
         self._n_link_dumps, self._n_processor_dumps, self._flags,
         self._link_proc_bits) = _PATTERN.unpack_from(data, offset)

    @property
    def router_wait1_timeout(self):
        """
        The WAIT1 timeout value of the router, in cycles.

        :rtype: int
        """
        return _decode_router_timeout_value(self._wait1_timeout)

    @property
    def router_wait1_timeout_parameters(self):
        """
        The WAIT1 timeout value of the router as mantissa and exponent.

        :rtype: tuple(int,int)
        """
        mantissa = self._wait1_timeout & 0xF
        exponent = (self._wait1_timeout >> 4) & 0xF
        return mantissa, exponent

    @property
    def router_wait2_timeout(self):
        """
        The WAIT2 timeout value of the router, in cycles.

        :rtype: int
        """
        return _decode_router_timeout_value(self._wait2_timeout)

    @property
    def router_wait2_timeout_parameters(self):
        """
        The WAIT2 timeout value of the router as mantissa and exponent.

        :rtype: tuple(int,int)
        """
        mantissa = self._wait2_timeout & 0xF
        exponent = (self._wait2_timeout >> 4) & 0xF
        return mantissa, exponent

    @property
    def n_dropped_packets(self):
        """
        The number of packets dropped by the router and received by
        the reinjection functionality (may not fit in the queue though).

        :rtype: int
        """
        return self._n_dropped_packets

    @property
    def n_missed_dropped_packets(self):
        """
        The number of times that when a dropped packet was read it was
        found that another one or more packets had also been dropped,
        but had been missed.

        :rtype: int
        """
        return self._n_missed_dropped_packets

    @property
    def n_dropped_packet_overflows(self):
        """
        Of the n_dropped_packets received, how many were lost due to not
        having enough space in the queue of packets to reinject.

        :rtype: int
        """
        return self._n_dropped_packet_overflows

    @property
    def n_processor_dumps(self):
        """
        The number of times that when a dropped packet was caused due to
        a processor failing to take the packet.

        :rtype: int
        """
        return self._n_processor_dumps

    @property
    def n_link_dumps(self):
        """
        The number of times that when a dropped packet was caused due to
        a link failing to take the packet.

        :rtype: int
        """
        return self._n_link_dumps

    @property
    def n_reinjected_packets(self):
        """
        Of the n_dropped_packets received, how many packets were
        successfully re-injected.

        :rtype: int
        """
        return self._n_reinjected_packets

    def _flag_set(self, flag):
        return (self._flags & flag.value) != 0

    @property
    def is_reinjecting_multicast(self):
        """
        True if re-injection of multicast packets is enabled.

        :rtype: bool
        """
        return self._flag_set(DPRIFlags.MULTICAST)

    @property
    def is_reinjecting_point_to_point(self):
        """
        True if re-injection of point-to-point packets is enabled.

        :rtype: bool
        """
        return self._flag_set(DPRIFlags.POINT_TO_POINT)

    @property
    def is_reinjecting_nearest_neighbour(self):
        """
        True if re-injection of nearest neighbour packets is enabled.

        :rtype: bool
        """
        return self._flag_set(DPRIFlags.NEAREST_NEIGHBOUR)

    @property
    def is_reinjecting_fixed_route(self):
        """
        True if re-injection of fixed-route packets is enabled.

        :rtype: bool
        """
        return self._flag_set(DPRIFlags.FIXED_ROUTE)

    @property
    def links_dropped_from(self):
        return [
            link for link in range(6) if self._link_proc_bits & (1 << link)]

    @property
    def processors_dropped_from(self):
        return [
            p for p in range(18) if self._link_proc_bits & (1 << p + 6)]



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .application_finisher import application_finisher
from .application_runner import application_runner
from .chip_iobuf_clearer import chip_io_buf_clearer
from .chip_iobuf_extractor import chip_io_buf_extractor
from .chip_provenance_updater import chip_provenance_updater
from .chip_runtime_updater import chip_runtime_updater
from .compute_energy_used import compute_energy_used
from .database_interface import database_interface
from .system_multicast_routing_generator import (
    system_multicast_routing_generator)
from .dsg_region_reloader import reload_dsg_regions
from .energy_provenance_reporter import energy_provenance_reporter
from .find_application_chips_used import FindApplicationChipsUsed
from .graph_binary_gatherer import graph_binary_gatherer
from .graph_data_specification_writer import (
    graph_data_specification_writer)
from .graph_provenance_gatherer import graph_provenance_gatherer
from .hbp_allocator import hbp_allocator
from .host_bit_field_router_compressor import (
    host_based_bit_field_router_compressor)
from .host_execute_data_specification import (
    execute_application_data_specs, execute_system_data_specs)
from .insert_chip_power_monitors_to_graphs import (
    insert_chip_power_monitors_to_graphs)
from .insert_extra_monitor_vertices_to_graphs import (
    insert_extra_monitor_vertices_to_graphs)
from .split_lpg_vertices import split_lpg_vertices
from .load_executable_images import load_app_images, load_sys_images
from .load_fixed_routes import load_fixed_routes
from .local_tdma_builder import local_tdma_builder
from .locate_executable_start_type import locate_executable_start_type
from .machine_generator import machine_generator
from .create_notification_protocol import create_notification_protocol
from .placements_provenance_gatherer import placements_provenance_gatherer
from .profile_data_gatherer import profile_data_gatherer
from .router_provenance_gatherer import router_provenance_gatherer
from .routing_setup import routing_setup
from .routing_table_loader import routing_table_loader
from .spalloc_allocator import spalloc_allocator
from .tags_loader import tags_loader
from .virtual_machine_generator import virtual_machine_generator
from .read_routing_tables_from_machine import read_routing_tables_from_machine
from .sdram_outgoing_partition_allocator import (
    sdram_outgoing_partition_allocator)
from .command_sender_adder import add_command_senders


__all__ = [
    "application_finisher", "application_runner", "chip_io_buf_clearer",
    "chip_io_buf_extractor", "chip_provenance_updater",
    "chip_runtime_updater", "create_notification_protocol",
    "compute_energy_used", "database_interface",
    "reload_dsg_regions",
    "energy_provenance_reporter", "execute_application_data_specs",
    "execute_system_data_specs", "FindApplicationChipsUsed",
    "graph_binary_gatherer", "graph_data_specification_writer",
    "graph_provenance_gatherer",
    "hbp_allocator", "host_based_bit_field_router_compressor",
    "insert_chip_power_monitors_to_graphs",
    "insert_extra_monitor_vertices_to_graphs",
    "split_lpg_vertices",
    "load_app_images", "load_fixed_routes", "load_sys_images",
    "local_tdma_builder", "locate_executable_start_type",
    "machine_generator", "placements_provenance_gatherer",
    "profile_data_gatherer",
    "read_routing_tables_from_machine", "router_provenance_gatherer",
    "routing_setup",
    "routing_table_loader", "sdram_outgoing_partition_allocator",
    "spalloc_allocator",
    "system_multicast_routing_generator", "tags_loader",
    "virtual_machine_generator", "add_command_senders"
]



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.data import FecDataView


def routing_table_loader(router_tables):
    """
    Loads routes into initialised routers.

    :param ~pacman.model.routing_tables.MulticastRoutingTables router_tables:
    """
    progress = ProgressBar(router_tables.routing_tables,
                           "Loading routing data onto the machine")

    # load each router table that is needed for the application to run into
    # the chips SDRAM
    app_id = FecDataView.get_app_id()
    transceiver = FecDataView.get_transceiver()
    for table in progress.over(router_tables.routing_tables):
        if (table.number_of_entries):
            transceiver.load_multicast_routes(
                table.x, table.y, table.multicast_routing_entries,
                app_id)



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import requests
from spinn_utilities.config_holder import get_config_str
from spinn_utilities.overrides import overrides
from spinn_front_end_common.abstract_models.impl import (
    MachineAllocationController)
from spinn_front_end_common.abstract_models import (
    AbstractMachineAllocationController)
from spinn_front_end_common.data import FecDataView
from pacman.exceptions import PacmanConfigurationException


class _HBPJobController(MachineAllocationController):
    __slots__ = [
        # the URLs to call the HBP system
        "_extend_lease_url",
        "_check_lease_url",
        "_release_machine_url",
        "_set_power_url",
        "_where_is_url",
        "_machine_name",
        "_power_on"
    ]

    _WAIT_TIME_MS = 10000

    def __init__(self, url, machine_name):
        """
        :param str url:
        :param str machine_name:
        """
        self._extend_lease_url = f"{url}/extendLease"
        self._check_lease_url = f"{url}/checkLease"
        self._release_machine_url = url
        self._set_power_url = f"{url}/power"
        self._where_is_url = f"{url}/chipCoordinates"
        self._machine_name = machine_name
        self._power_on = True
        # Lower the level of requests to WARNING to avoid extra messages
        logging.getLogger("requests").setLevel(logging.WARNING)
        super().__init__("HBPJobController")

    @overrides(AbstractMachineAllocationController.extend_allocation)
    def extend_allocation(self, new_total_run_time):
        r = requests.get(self._extend_lease_url, params={
            "runTime": new_total_run_time}, timeout=10)
        r.raise_for_status()

    def _check_lease(self, wait_time):
        r = requests.get(self._check_lease_url, params={
            "waitTime": wait_time}, timeout=10)
        r.raise_for_status()
        return r.json()

    def _release(self, machine_name):
        r = requests.delete(self._release_machine_url, params={
            "machineName": machine_name}, timeout=10)
        r.raise_for_status()

    def _set_power(self, machine_name, power_on):
        r = requests.put(self._set_power_url, params={
            "machineName": machine_name, "on": bool(power_on)}, timeout=10)
        r.raise_for_status()

    def _where_is(self, machine_name, chip_x, chip_y):
        r = requests.get(self._where_is_url, params={
            "machineName": machine_name, "chipX": chip_x,
            "chipY": chip_y}, timeout=10)
        r.raise_for_status()
        return r.json()

    @overrides(AbstractMachineAllocationController.close)
    def close(self):
        super().close()
        self._release(self._machine_name)

    @overrides(MachineAllocationController._teardown)
    def _teardown(self):
        self._release(self._machine_name)

    @property
    def power(self):
        return self._power_on

    def set_power(self, power):
        self._set_power(self._machine_name, power)
        self._power_on = power

    @overrides(AbstractMachineAllocationController.where_is_machine)
    def where_is_machine(self, chip_x, chip_y):
        return self._where_is(self._machine_name, chip_x, chip_y)

    @overrides(MachineAllocationController._wait)
    def _wait(self):
        return self._check_lease(self._WAIT_TIME_MS)["allocated"]


def hbp_allocator(total_run_time):
    """
    Request a machine from the HBP remote access server that will fit
    a number of chips.

    :param int total_run_time: The total run time to request
    :return: machine name, machine version, BMP details (if any),
        reset on startup flag, auto-detect BMP, SCAMP connection details,
        boot port, allocation controller
    :rtype: tuple(str, int, object, bool, bool, object, object,
        MachineAllocationController)
    :raises ~pacman.exceptions.PacmanConfigurationException:
        If neither `n_chips` or `n_boards` provided
    """

    url = str(get_config_str("Machine", "remote_spinnaker_url"))
    if url.endswith("/"):
        url = url[:-1]

    machine = _get_machine(url, total_run_time)
    hbp_job_controller = _HBPJobController(url, machine["machineName"])

    bmp_details = None
    if "bmp_details" in machine:
        bmp_details = machine["bmpDetails"]

    return (
        machine["machineName"], int(machine["version"]),
        bmp_details, False, False, None, hbp_job_controller)


def _get_machine(url, total_run_time):
    """
    :param str url:
    :param int total_run_time:
    :rtype: dict
    """
    if FecDataView.has_n_boards_required():
        get_machine_request = requests.get(
            url, params={"nBoards": FecDataView.get_n_boards_required(),
                         "runTime": total_run_time}, timeout=10)
    elif FecDataView.has_n_chips_needed():
        get_machine_request = requests.get(
            url, params={"nChips": FecDataView.get_n_chips_needed(),
                         "runTime": total_run_time}, timeout=10)
    else:
        raise PacmanConfigurationException(
            "At least one of n_chips or n_boards must be provided")

    get_machine_request.raise_for_status()
    return get_machine_request.json()



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import defaultdict
from spinn_machine.router import Router
from spinn_front_end_common.data import FecDataView

COLLISION_REPORT = "routing_collision_protential_report.rpt"


def router_collision_potential_report():
    file_name = os.path.join(
        FecDataView.get_run_dir_path(),
        "routing_collision_protential_report.rpt")
    with open(file_name, "w", encoding="utf-8") as writer:
        collision_counts = _generate_data()
        _write_report(collision_counts, writer)


def _write_report(collision_counts, writer):
    if len(collision_counts) == 0:
        writer.write("There are no collisions in this network mapping")
        return

    for (x, y) in collision_counts:
        for link_id in collision_counts[(x, y)]:
            writer.write(
                f"router {x}:{y} link {link_id} has potential "
                f"{collision_counts[(x, y)][link_id]} collisions\n")


def _generate_data():
    router_tables_by_partition = FecDataView.get_routing_table_by_partition()
    collisions = defaultdict(lambda: defaultdict(int))
    n_keys_map = dict()
    # https://github.com/SpiNNakerManchester/SpiNNFrontEndCommon/issues/891
    # use machine_vertex.get_n_keys_for_partition(identifier)
    for (x, y) in router_tables_by_partition.get_routers():
        for partition in \
                router_tables_by_partition.get_entries_for_router(x, y):
            entry = router_tables_by_partition.get_entry_on_coords_for_edge(
                partition, x, y)
            for link in entry.link_ids:
                if collisions[(x, y)][link] == 0:
                    chip = FecDataView.get_chip_at(x, y)
                    other_chip_x = chip.router.get_link(link).destination_x
                    other_chip_y = chip.router.get_link(link).destination_y
                    collision_route = Router.opposite(link)

                    collision_potential = _get_collisions_with_other_router(
                        other_chip_x, other_chip_y, collision_route,
                        router_tables_by_partition, n_keys_map)

                    collisions[(x, y)][link] = collision_potential

                n_packets = 0  # n_keys_map.n_keys_for_partition(partition)
                collisions[(x, y)][link] += n_packets
    return collisions


def _get_collisions_with_other_router(
        x, y, collision_route, router_tables_by_partition, n_key_map):
    total = 0
    for partition in router_tables_by_partition.get_entries_for_router(x, y):
        entry = router_tables_by_partition.get_entry_on_coords_for_edge(
            partition, x, y)
        if collision_route in entry.link_ids:
            total += n_key_map.n_keys_for_partition(partition)
    return total



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pacman.utilities.utility_calls import md5
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.constants import SDP_PORTS


def get_simulation_header_array(binary_file_name):
    """
    Get data to be written to the simulation header.

    :param str binary_file_name: The name of the binary of the application
    :return: An array of values to be written as the simulation header
    :rtype: list(int)
    """
    # Get first 32-bits of the md5 hash of the application name
    application_name_hash = md5(os.path.splitext(binary_file_name)[0])[:8]

    # Write this to the system region (to be picked up by the simulation):
    data = list()
    data.append(int(application_name_hash, 16))
    data.append(FecDataView.get_hardware_time_step_us())

    # add SDP port number for receiving synchronisations and new run times
    data.append(SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value)

    return data


def get_simulation_header_array_no_timestep(binary_file_name):
    """
    Get data to be written to the simulation header.

    :param str binary_file_name: The name of the binary of the application
    :return: An array of values to be written as the simulation header
    :rtype: list(int)
    """
    # Get first 32-bits of the md5 hash of the application name
    application_name_hash = md5(os.path.splitext(binary_file_name)[0])[:8]

    # Write this to the system region (to be picked up by the simulation):
    data = list()
    data.append(int(application_name_hash, 16))
    data.append(0)

    # add SDP port number for receiving synchronisations and new run times
    data.append(SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value)

    return data



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import struct
from spinn_utilities.log import FormatAdapter
from spinn_machine import CoreSubsets
from spinnman.model.enums import CPUState
from data_specification.constants import (
    APP_PTR_TABLE_HEADER_BYTE_SIZE, APP_PTR_TABLE_REGION_BYTE_SIZE)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.utility_objs import ExecutableType

logger = FormatAdapter(logging.getLogger(__name__))
_n_word_structs = []


def locate_extra_monitor_mc_receiver(placement_x, placement_y):
    """
    Get the data speed up gatherer that can be used to talk to a
    particular chip. This will be on the same board.

    :param int placement_x: The X coordinate of the reference chip
    :param int placement_y: The Y coordinate of the reference chip
    :rtype: DataSpeedUpPacketGatherMachineVertex
    """
    chip = FecDataView.get_chip_at(placement_x, placement_y)
    return FecDataView.get_gatherer_by_xy(
        chip.nearest_ethernet_x, chip.nearest_ethernet_y)


def read_data(x, y, address, length, data_format):
    """
    Reads and converts a single data item from memory.

    :param int x: chip x
    :param int y: chip y
    :param int address: base address of the SDRAM chip to read
    :param int length: length to read
    :param str data_format:
        the format to read memory (see :py:func:`struct.pack`)
    :return: whatever is produced by unpacking the data
    :rtype: tuple
    """
    # pylint: disable=too-many-arguments
    data = FecDataView.read_memory(x, y, address, length)
    return struct.unpack_from(data_format, data)[0]


def write_address_to_user0(x, y, p, address):
    """
    Writes the given address into the user_0 register of the given core.

    :param int x: Chip coordinate.
    :param int y: Chip coordinate.
    :param int p: Core ID on chip.
    :param int address: Value to write (32-bit integer)
    """
    txrx = FecDataView.get_transceiver()
    user_0_address = txrx.get_user_0_register_address_from_core(p)
    txrx.write_memory(x, y, user_0_address, address)


def write_address_to_user1(x, y, p, address):
    """
    Writes the given address into the user_1 register of the given core.

    :param int x: Chip coordinate.
    :param int y: Chip coordinate.
    :param int p: Core ID on chip.
    :param int address: Value to write (32-bit integer)
    """
    txrx = FecDataView.get_transceiver()
    user_1_address = txrx.get_user_1_register_address_from_core(p)
    txrx.write_memory(x, y, user_1_address, address)


def get_region_base_address_offset(app_data_base_address, region):
    """
    Find the address of the of a given region for the DSG.

    :param int app_data_base_address: base address for the core
    :param int region: the region ID we're looking for
    """
    return (app_data_base_address +
            APP_PTR_TABLE_HEADER_BYTE_SIZE +
            (region * APP_PTR_TABLE_REGION_BYTE_SIZE))


def locate_memory_region_for_placement(placement, region):
    """
    Get the address of a region for a placement.

    :param int region: the region to locate the base address of
    :param ~pacman.model.placements.Placement placement:
        the placement object to get the region address of
    :param ~spinnman.transceiver.Transceiver transceiver:
        the python interface to the SpiNNaker machine
    :return: the address
    :rtype: int
    """
    transceiver = FecDataView.get_transceiver()
    regions_base_address = transceiver.get_cpu_information_from_core(
        placement.x, placement.y, placement.p).user[0]

    # Get the position of the region in the pointer table
    element_addr = get_region_base_address_offset(regions_base_address, region)

    # Get the actual address of the region
    return transceiver.read_word(placement.x, placement.y, element_addr)


def convert_string_into_chip_and_core_subset(cores):
    """
    Translate a string list of cores into a core subset.

    :param cores:
        string representing down cores formatted as x,y,p[:x,y,p]*
    :type cores: str or None
    :rtype: ~spinn_machine.CoreSubsets
    """
    ignored_cores = CoreSubsets()
    if cores is not None and cores != "None":
        for downed_core in cores.split(":"):
            x, y, processor_id = downed_core.split(",")
            ignored_cores.add_processor(int(x), int(y), int(processor_id))
    return ignored_cores


def flood_fill_binary_to_spinnaker(binary):
    """
    Flood fills a binary to SpiNNaker.

    :param str binary:
        The name of the file containing the APLX binary to load
    :return: the number of cores it was loaded onto
    :rtype: int
    """
    executable_targets = FecDataView.get_executable_targets()
    core_subset = executable_targets.get_cores_for_binary(binary)
    FecDataView.get_transceiver().execute_flood(
        core_subset, binary, FecDataView.get_app_id(), wait=True,
        is_filename=True)
    return len(core_subset)


def generate_unique_folder_name(folder, filename, extension):
    """
    Generate a unique file name with a given extension in a given folder.

    :param str folder: where to put this unique file
    :param str filename:
        the name of the first part of the file without extension
    :param str extension: extension of the file
    :return: file path with a unique addition
    :rtype: str
    """
    new_file_path = os.path.join(folder, f"{filename}{extension}")
    count = 2
    while os.path.exists(new_file_path):
        new_file_path = os.path.join(folder, f"{filename}_{count}{extension}")
        count += 1
    return new_file_path


def get_ethernet_chip(machine, board_address):
    """
    Locate the chip with the given board IP address.

    :param ~spinn_machine.Machine machine: the SpiNNaker machine
    :param str board_address: the board address to locate the chip of.
    :return: The chip that supports that board address
    :rtype: ~spinn_machine.Chip
    :raises ConfigurationException:
        when that board address has no chip associated with it
    """
    for chip in machine.ethernet_connected_chips:
        if chip.ip_address == board_address:
            return chip
    raise ConfigurationException(
        "cannot find the Ethernet connected chip with the "
        f"board address {board_address}")


def determine_flow_states(executable_types, no_sync_changes):
    """
    Get the start and end states for these executable types.

    :param dict(ExecutableType,any) executable_types:
        the execute types to locate start and end states from
    :param int no_sync_changes: the number of times sync signals been sent
    :return: dict of executable type to states.
    :rtype: tuple(dict(ExecutableType,~spinnman.model.enums.CPUState),
        dict(ExecutableType,~spinnman.model.enums.CPUState))
    """
    expected_start_states = dict()
    expected_end_states = dict()
    for start_type in executable_types.keys():

        # cores that ignore all control and are just running
        if start_type == ExecutableType.RUNNING:
            expected_start_states[ExecutableType.RUNNING] = [
                CPUState.RUNNING, CPUState.FINISHED]
            expected_end_states[ExecutableType.RUNNING] = [
                CPUState.RUNNING, CPUState.FINISHED]

        # cores that require a sync barrier
        elif start_type == ExecutableType.SYNC:
            expected_start_states[ExecutableType.SYNC] = [CPUState.SYNC0]
            expected_end_states[ExecutableType.SYNC] = [CPUState.FINISHED]

        # cores that use our sim interface
        elif start_type == ExecutableType.USES_SIMULATION_INTERFACE:
            if no_sync_changes % 2 == 0:
                expected_start_states[start_type] = [CPUState.SYNC0]
            else:
                expected_start_states[start_type] = [CPUState.SYNC1]
            expected_end_states[start_type] = [CPUState.PAUSED]

    # if no states, go boom.
    if not expected_start_states:
        raise ConfigurationException(
            f"Unknown executable start types {executable_types}")
    return expected_start_states, expected_end_states


def convert_vertices_to_core_subset(vertices):
    """
    Converts vertices into core subsets.

    :param iterable(~pacman.model.graphs.machine.MachineVertex) vertices:
        the vertices to convert to core subsets
    :return: the CoreSubSets of the vertices
    :rtype: ~spinn_machine.CoreSubsets
    """
    core_subsets = CoreSubsets()
    for vertex in vertices:
        placement = FecDataView.get_placement_of_vertex(vertex)
        core_subsets.add_processor(placement.x, placement.y, placement.p)
    return core_subsets


def n_word_struct(n_words):
    """
    Manages a precompiled cache of :py:class`~struct.Struct`\\s for
    parsing blocks of words.
    Thus, this::

        data = n_word_struct(n_words).unpack(data_blob)

    Is much like doing this::

        data = struct.unpack(f"<{n_words}I", data_blob)

    except quite a bit more efficient because things are shared including the
    cost of parsing the format.

    :param int n_words: The number of *SpiNNaker words* to be handled.
    :return: A struct for working with that many words.
    :rtype: ~struct.Struct
    """
    # pylint: disable=global-statement
    global _n_word_structs
    while len(_n_word_structs) < n_words + 1:
        _n_word_structs += [None] * (n_words + 1 - len(_n_word_structs))
    s = _n_word_structs[n_words]
    if s is not None:
        return s
    new_struct = struct.Struct(f"<{n_words}I")
    _n_word_structs[n_words] = new_struct
    return new_struct


def get_defaultable_source_id(entry):
    """
    Hack to support the source requirement for the router compressor on chip.

    :param ~spinn_machine.MulticastRoutingEntry entry:
        the multicast router table entry.
    :return: return the source value
    :rtype: int
    """
    if entry.defaultable:
        return (list(entry.link_ids)[0] + 3) % 6
    elif entry.link_ids:
        return list(entry.link_ids)[0]
    return 0



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.data import FecDataView


def load_fixed_routes():
    """
    Load a set of fixed routes onto a SpiNNaker machine.

    :param ~spinnman.transceiver.Transceiver transceiver:
    """
    fixed_routes = FecDataView.get_fixed_routes()
    progress_bar = ProgressBar(
        total_number_of_things_to_do=len(fixed_routes),
        string_describing_what_being_progressed="loading fixed routes")
    app_id = FecDataView.get_app_id()
    transceiver = FecDataView.get_transceiver()
    for chip_x, chip_y in progress_bar.over(fixed_routes.keys()):
        transceiver.load_fixed_route(
            chip_x, chip_y, fixed_routes[chip_x, chip_y], app_id)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.utilities.database import DatabaseWriter
from spinn_front_end_common.abstract_models import (
    AbstractSupportsDatabaseInjection)
from spinn_front_end_common.data import FecDataView

logger = FormatAdapter(logging.getLogger(__name__))


def database_interface(runtime):
    """
    :param int runtime:
    :return: where the database is located
    :rtype: str
    """
    # pylint: disable=too-many-arguments
    needs_db = DatabaseWriter.auto_detect_database()
    user_create_database = get_config_bool("Database", "create_database")
    if user_create_database is not None:
        if user_create_database != needs_db:
            logger.warning(f"Database creating changed to "
                           f"{user_create_database} due to cfg settings")
            needs_db = user_create_database

    if needs_db:
        writer = DatabaseWriter()
        logger.info("Creating live event connection database in {}",
                    writer.database_path)
        _write_to_db(writer, runtime)
        writer.close()
        return writer.database_path
    return None


def _write_to_db(writer, runtime):
    """
    :param DatabaseWriter writer:
    :param int runtime:
    """
    with writer as w, ProgressBar(
            6, "Creating graph description database") as p:
        w.add_system_params(runtime)
        w.add_proxy_configuration()
        p.update()
        w.add_machine_objects()
        p.update()
        w.add_application_vertices()
        p.update()
        w.add_placements()
        p.update()
        w.add_tags()
        p.update()
        lpg_source_machine_vertices = w.add_lpg_mapping()

        if get_config_bool(
                "Database", "create_routing_info_to_neuron_id_mapping"):
            machine_vertices = {
                (vertex, vertex.injection_partition_id)
                for vertex in FecDataView.iterate_machine_vertices()
                if isinstance(vertex, AbstractSupportsDatabaseInjection)
                and vertex.is_in_injection_mode}
            machine_vertices.update(lpg_source_machine_vertices)
            live_vertices = FecDataView.iterate_live_output_vertices()
            machine_vertices.update(
                (m_vertex, part_id)
                for vertex, part_id in live_vertices
                for m_vertex in vertex.splitter.get_out_going_vertices(
                    part_id))
            w.create_atom_to_event_id_mapping(machine_vertices)
        p.update()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from spinn_utilities.config_holder import (get_config_int, get_config_str)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.provenance import (
    GlobalProvenance, ProvenanceReader, TimerCategory, TimerWork)
from spinn_front_end_common.utilities.utility_objs import PowerUsed
from spinn_front_end_common.utility_models import (
    ChipPowerMonitorMachineVertex)

#: milliseconds per second
_MS_PER_SECOND = 1000.0

#: given from Indar's measurements
MILLIWATTS_PER_FPGA = 0.000584635

#: stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
#: Massively-Parallel Neural Network Simulation)
JOULES_PER_SPIKE = 0.000000000800

#: stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
#: Massively-Parallel Neural Network Simulation)
MILLIWATTS_PER_IDLE_CHIP = 0.000360

#: stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
#: Massively-Parallel Neural Network Simulation)
MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD = (0.001 - MILLIWATTS_PER_IDLE_CHIP)

#: measured from the real power meter and timing between
#: the photos for a days powered off
MILLIWATTS_FOR_FRAME_IDLE_COST = 0.117

#: measured from the loading of the column and extrapolated
MILLIWATTS_PER_FRAME_ACTIVE_COST = 0.154163558

#: measured from the real power meter and timing between the photos
#: for a day powered off
MILLIWATTS_FOR_BOXED_48_CHIP_FRAME_IDLE_COST = 0.0045833333

# TODO needs filling in
MILLIWATTS_PER_UNBOXED_48_CHIP_FRAME_IDLE_COST = 0.01666667

# TODO verify this is correct when doing multiboard comms
N_MONITORS_ACTIVE_DURING_COMMS = 2


def compute_energy_used():
    """
    This algorithm does the actual work of computing energy used by a
    simulation (or other application) running on SpiNNaker.

    :rtype: PowerUsed
    """
    machine = FecDataView.get_machine()
    with GlobalProvenance() as db:
        dsg_time = db.get_category_timer_sum(TimerCategory.DATA_GENERATION)
        execute_time = db.get_category_timer_sum(TimerCategory.RUN_LOOP)
        # NOTE: this extraction time is part of the execution time; it does not
        # refer to the time taken in e.g. pop.get_data() or projection.get()
        extraction_time = db.get_timer_sum_by_work(TimerWork.EXTRACT_DATA)
        load_time = db.get_category_timer_sum(TimerCategory.LOADING)
        mapping_time = db.get_category_timer_sum(TimerCategory.MAPPING)
    # TODO get_machine not include here
    power_used = PowerUsed()

    power_used.num_chips = machine.n_chips
    # One extra per chip for SCAMP
    power_used.num_cores = FecDataView.get_n_placements() + machine.n_chips
    power_used.exec_time_secs = execute_time / _MS_PER_SECOND
    power_used.loading_time_secs = load_time / _MS_PER_SECOND
    # extraction_time could be None if nothing is set to be recorded
    total_extraction_time = 0
    if extraction_time is not None:
        total_extraction_time += extraction_time
    power_used.saving_time_secs = total_extraction_time / _MS_PER_SECOND
    power_used.data_gen_time_secs = dsg_time / _MS_PER_SECOND
    power_used.mapping_time_secs = mapping_time / _MS_PER_SECOND

    runtime_total_ms = (
            FecDataView.get_current_run_timesteps() *
            FecDataView.get_time_scale_factor())
    # TODO: extraction time as currently defined is part of execution time,
    #       so for now don't add it on, but revisit this in the future
    total_booted_time = execute_time + load_time
    _compute_energy_consumption(
         machine, dsg_time, load_time, mapping_time, total_booted_time,
         runtime_total_ms, power_used)

    return power_used


def _compute_energy_consumption(
        machine, dsg_time, load_time, mapping_time, total_booted_time,
        runtime_total_ms, power_used):
    """
    :param ~.Machine machine:
    :param float dsg_time:
    :param float load_time:
    :param float mapping_time:
    :param float total_booted_time:
    :param float runtime_total_ms:
    :param PowerUsed power_used:
    """
    # figure active chips
    monitor_placements = __find_monitor_placements()

    # figure out packet cost
    _router_packet_energy(power_used)

    # figure FPGA cost over all booted and during runtime cost
    _calculate_fpga_energy(
        machine, total_booted_time, runtime_total_ms, power_used)

    # figure how many frames are using, as this is a constant cost of
    # routers, cooling etc
    power_used.num_frames = _calculate_n_frames(machine)

    # figure load time cost
    power_used.loading_joules = _calculate_loading_energy(
        machine, load_time, len(monitor_placements), power_used.num_frames)

    # figure the down time idle cost for mapping
    power_used.mapping_joules = _calculate_power_down_energy(
        mapping_time, machine, power_used.num_frames)

    # figure the down time idle cost for DSG
    power_used.data_gen_joules = _calculate_power_down_energy(
        dsg_time, machine, power_used.num_frames)

    # figure extraction time cost
    power_used.saving_joules = _calculate_data_extraction_energy(
        machine, len(monitor_placements), power_used.num_frames)

    # figure out active chips cost
    power_used.chip_energy_joules = sum(
        _calculate_chips_active_energy(
            placement, runtime_total_ms, power_used)
        for placement in monitor_placements)

    # figure out cooling/internet router idle cost during runtime
    power_used.baseline_joules = (
        runtime_total_ms * power_used.num_frames *
        MILLIWATTS_FOR_FRAME_IDLE_COST)


def __find_monitor_placements():
    return [placement
            for placement in FecDataView.iterate_placements_by_vertex_type(
                ChipPowerMonitorMachineVertex)]


_COST_PER_TYPE = {
    "Local_Multicast_Packets": JOULES_PER_SPIKE,
    "External_Multicast_Packets": JOULES_PER_SPIKE,
    "Reinjected": JOULES_PER_SPIKE,
    "Local_P2P_Packets": JOULES_PER_SPIKE * 2,
    "External_P2P_Packets": JOULES_PER_SPIKE * 2,
    "Local_NN_Packets": JOULES_PER_SPIKE,
    "External_NN_Packets": JOULES_PER_SPIKE,
    "Local_FR_Packets": JOULES_PER_SPIKE * 2,
    "External_FR_Packets": JOULES_PER_SPIKE * 2
}


def _router_packet_energy(power_used):
    """
    :param PowerUsed power_used:
    """
    energy_cost = 0.0
    with ProvenanceReader() as db:
        for name, cost in _COST_PER_TYPE.items():
            data = db.get_router_by_chip(name)
            for (x, y, value) in data:
                this_cost = value * cost
                energy_cost += this_cost
                if this_cost:
                    power_used.add_router_active_energy(x, y, this_cost)

    power_used.packet_joules = energy_cost


def _calculate_chips_active_energy(placement, runtime_total_ms, power_used):
    """
    Figure out the chip active cost during simulation.

    :param ~.Placement placement: placement
    :param float runtime_total_ms:
    :param PowerUsed power_used:
    :return: energy cost
    """
    # locate chip power monitor
    chip_power_monitor = placement.vertex

    # get recordings from the chip power monitor
    recorded_measurements = chip_power_monitor.get_recorded_data(
        placement=placement)

    # deduce time in milliseconds per recording element
    n_samples_per_recording = get_config_int(
        "EnergyMonitor", "n_samples_per_recording_entry")
    time_for_recorded_sample = (
        chip_power_monitor.sampling_frequency *
        n_samples_per_recording) / 1000
    cores_power_cost = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    # accumulate costs
    for recorded_measurement in recorded_measurements:
        for core in range(0, 18):
            cores_power_cost[core] += (
                recorded_measurement[core] * time_for_recorded_sample *
                MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD / 18)

    # detailed report print out
    for core in range(0, 18):
        power_used.add_core_active_energy(
            placement.x, placement.y, core, cores_power_cost[core])

    # TAKE INTO ACCOUNT IDLE COST
    idle_cost = runtime_total_ms * MILLIWATTS_PER_IDLE_CHIP
    return sum(cores_power_cost) + idle_cost


def _calculate_fpga_energy(
        machine, total_runtime, runtime_total_ms, power_used):
    """
    :param ~.Machine machine:
    :param float total_runtime:
    :param float runtime_total_ms:
    :param PowerUsed power_used:
    """
    total_fpgas = 0
    # if not spalloc, then could be any type of board
    if (not get_config_str("Machine", "spalloc_server") and
            not get_config_str("Machine", "remote_spinnaker_url")):
        # if a spinn2 or spinn3 (4 chip boards) then they have no fpgas
        if machine.n_chips <= 4:
            return 0, 0

        # if the spinn4 or spinn5 board, need to verify if wrap-arounds
        # are there, if not then assume fpgas are turned off.

        # how many fpgas are active
        total_fpgas = __board_n_operational_fpgas(
            machine.ethernet_connected_chips[0])
        # active fpgas
        if total_fpgas == 0:
            return 0, 0
    else:  # spalloc machine, need to check each board
        for ethernet_connected_chip in machine.ethernet_connected_chips:
            total_fpgas += __board_n_operational_fpgas(
               ethernet_connected_chip)

    # Only need to update this here now that we've learned there are FPGAs
    # in use
    power_used.num_fpgas = total_fpgas
    power_usage_total = (
        total_runtime * MILLIWATTS_PER_FPGA * total_fpgas)
    power_usage_runtime = (
        runtime_total_ms * MILLIWATTS_PER_FPGA * total_fpgas)
    power_used.fpga_total_energy_joules = power_usage_total
    power_used.fpga_exec_energy_joules = power_usage_runtime


def __board_n_operational_fpgas(ethernet_chip):
    """
    Figures out how many FPGAs were switched on for a particular SpiNN-5 board.

    :param ~.Chip ethernet_chip: the Ethernet-enabled chip to look from
    :return: number of FPGAs on, on this board
    :rtype: int
    """
    # TODO: should be possible to get this info from Machine

    # As the Chips can be None use the machine call and not the View call
    machine = FecDataView.get_machine()

    # positions to check for active links
    left_chips = (
        machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
        for dx, dy in ((0, 0), (0, 1), (0, 2), (0, 3), (0, 4)))
    right_chips = (
        machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
        for dx, dy in ((7, 3), (7, 4), (7, 5), (7, 6), (7, 7)))
    top_chips = (
        machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
        for dx, dy in ((4, 7), (5, 7), (6, 7), (7, 7)))
    bottom_chips = (
        machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
        for dx, dy in ((0, 0), (1, 0), (2, 0), (3, 0), (4, 0)))
    top_left_chips = (
        machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
        for dx, dy in ((0, 3), (1, 4), (2, 5), (3, 6), (4, 7)))
    bottom_right_chips = (
        machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
        for dx, dy in ((0, 4), (1, 5), (2, 6), (3, 7)))

    # bottom left, bottom
    fpga_0 = __deduce_fpga(
        bottom_chips, bottom_right_chips, (5, 4), (0, 5))
    # left, and top right
    fpga_1 = __deduce_fpga(
        left_chips, top_left_chips, (3, 4), (3, 2))
    # top and right
    fpga_2 = __deduce_fpga(
        top_chips, right_chips, (2, 1), (0, 1))
    return fpga_0 + fpga_1 + fpga_2


def __deduce_fpga(chips_1, chips_2, links_1, links_2):
    """
    Figure out if each FPGA was on or not.

    :param iterable(~.Chip) chips_1: chips on an edge of the board
    :param iterable(~.Chip) chips_2: chips on an edge of the board
    :param iterable(int) links_1: which link IDs to check from chips_1
    :param iterable(int) links_2: which link IDs to check from chips_2
    :return: 0 if not on, 1 if on
    :rtype: int
    """
    for chip, link_id in itertools.product(chips_1, links_1):
        if chip and chip.router.get_link(link_id) is not None:
            return 1
    for chip, link_id in itertools.product(chips_2, links_2):
        if chip and chip.router.get_link(link_id) is not None:
            return 1
    return 0


def _calculate_loading_energy(machine, load_time_ms, n_monitors, n_frames):
    """
    :param ~.Machine machine:
    :param float load_time_ms: milliseconds
    :param int n_monitors
    :param int n_frames:
    :rtype: float
    """
    # find time in milliseconds
    with GlobalProvenance() as db:
        total_time_ms = db.get_timer_sum_by_category(TimerCategory.LOADING)

    # handle monitor core active cost

    # min between chips that are active and fixed monitor, as when 1
    # chip is used its one monitor, if more than 1 chip,
    # the ethernet connected chip and the monitor handling the read/write
    # this is checked by min
    n_monitors_active = min(N_MONITORS_ACTIVE_DURING_COMMS, n_monitors)
    energy_cost = (
        total_time_ms * n_monitors_active *
        MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD /
        machine.DEFAULT_MAX_CORES_PER_CHIP)

    # handle all idle cores
    energy_cost += _calculate_idle_cost(total_time_ms, machine)

    # handle time diff between load time and total load phase of ASB
    energy_cost += (
        (total_time_ms - load_time_ms) *
        machine.n_chips * MILLIWATTS_PER_IDLE_CHIP)

    # handle active routers etc
    active_router_cost = (
        load_time_ms * n_frames * MILLIWATTS_PER_FRAME_ACTIVE_COST)

    # accumulate
    energy_cost += active_router_cost
    return energy_cost


def _calculate_data_extraction_energy(machine, n_monitors, n_frames):
    """
    Data extraction cost.

    :param ~.Machine machine: machine description
    :param int n_monitors:
    :param int n_frames:
    :return: cost of data extraction, in Joules
    :rtype: float
    """
    # find time
    # TODO is this what was desired
    total_time_ms = 0
    with GlobalProvenance() as db:
        buffer_time_ms = db.get_timer_sum_by_work(TimerWork.EXTRACT_DATA)

    energy_cost = 0
    # NOTE: Buffer time could be None if nothing was set to record
    if buffer_time_ms is not None:
        total_time_ms += buffer_time_ms

        # min between chips that are active and fixed monitor, as when 1
        # chip is used its one monitor, if more than 1 chip,
        # the ethernet connected chip and the monitor handling the read/write
        # this is checked by min
        energy_cost = (
            total_time_ms *
            min(N_MONITORS_ACTIVE_DURING_COMMS, n_monitors) *
            MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD /
            machine.DEFAULT_MAX_CORES_PER_CHIP)

        # add idle chip cost
        energy_cost += _calculate_idle_cost(total_time_ms, machine)

        # handle active routers etc
        energy_cost_of_active_router = (
            total_time_ms * n_frames * MILLIWATTS_PER_FRAME_ACTIVE_COST)
        energy_cost += energy_cost_of_active_router

    return energy_cost


def _calculate_idle_cost(time, machine):
    """
    Calculate energy used by being idle.

    :param float time: time machine was idle, in milliseconds
    :param ~.Machine machine: machine description
    :return: cost, in joules
    :rtype: float
    """
    return (time * machine.total_available_user_cores *
            MILLIWATTS_PER_IDLE_CHIP /
            machine.DEFAULT_MAX_CORES_PER_CHIP)


def _calculate_power_down_energy(time, machine, n_frames):
    """
    Calculate power down costs.

    :param float time: time powered down, in milliseconds
    :param ~.Machine machine:
    :param int n_frames: number of frames used by this machine
    :return: energy in joules
    :rtype: float
    """
    # if spalloc or hbp
    if FecDataView.has_allocation_controller():
        return time * n_frames * MILLIWATTS_FOR_FRAME_IDLE_COST
    # if 4 chip
    elif machine.n_chips <= 4:
        return machine.n_chips * time * MILLIWATTS_PER_IDLE_CHIP
    # if 48 chip
    else:
        return time * MILLIWATTS_FOR_BOXED_48_CHIP_FRAME_IDLE_COST


def _calculate_n_frames(machine):
    """
    Figures out how many frames are being used in this setup.
    A key of cabinet,frame will be used to identify unique frame.

    :param ~.Machine machine: the machine object
    :return: number of frames
    :rtype: int
    """
    # if not spalloc, then could be any type of board, but unknown cooling
    if not FecDataView.has_allocation_controller():
        return 0

    # using spalloc in some form; how many unique frames?
    cabinet_frame = set()
    mac = FecDataView.get_allocation_controller()
    for ethernet_connected_chip in machine.ethernet_connected_chips:
        cabinet, frame, _ = mac.where_is_machine(
            ethernet_connected_chip.x, ethernet_connected_chip.y)
        cabinet_frame.add((cabinet, frame))
    return len(cabinet_frame)



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import auto, Enum


class TimerCategory(Enum):
    """
    Different Categories a FecTimer can be in

    """
    # Category Constants
    WAITING = (auto(), "Waiting")
    SETTING_UP = (auto(), "In Setup")
    RUN_OTHER = (auto(), "In run other")
    GET_MACHINE = (auto(), "Turning on Machine")
    LOADING = (auto(), "Loading Stage")
    DATA_GENERATION = (auto(), "data_generation")
    MAPPING = (auto(), "Mapping Stage")
    RUN_LOOP = (auto(), "Running Stage")
    RESETTING = (auto(), "Resetting")
    SHUTTING_DOWN = (auto(), "Shutting down")

    def __new__(cls, value, category_name):
        # pylint: disable=protected-access
        obj = object.__new__(cls)
        obj._value_ = value
        obj._category_name = category_name
        return obj

    @property
    def category_name(self):
        return self._category_name



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.constants import SCP_SCAMP_PORT


class AbstractMachineAllocationController(object, metaclass=AbstractBase):
    """
    An object that controls the allocation of a machine
    """

    __slots__ = ()

    @abstractmethod
    def extend_allocation(self, new_total_run_time):
        """
        Extend the allocation of the machine from the original run time.

        :param float new_total_run_time:
            The total run time that is now required starting from when the
            machine was first allocated
        """

    @abstractmethod
    def close(self):
        """
        Indicate that the use of the machine is complete.
        """

    @abstractmethod
    def where_is_machine(self, chip_x, chip_y):
        """
        Locates and returns cabinet, frame, board for a given chip in a
        machine allocated to this job.

        :param int chip_x: chip x location
        :param int chip_y: chip y location
        :return: (cabinet, frame, board)
        :rtype: tuple(int,int,int)
        """

    @abstractmethod
    def create_transceiver(self):
        """
        Create a transceiver for talking to the allocated machine, and
        make sure everything is ready for use (i.e. boot and discover
        connections if needed).

        :rtype: ~spinnman.transceiver.Transceiver
        """

    @abstractmethod
    def open_sdp_connection(self, chip_x, chip_y, udp_port=SCP_SCAMP_PORT):
        """
        Open a connection to a specific Ethernet-enabled SpiNNaker chip.
        Caller will have to arrange for SpiNNaker to pay attention to the
        connection.

        The coordinates will be job-relative.

        :param int chip_x: Ethernet-enabled chip X coordinate
        :param int chip_y: Ethernet-enabled chip Y coordinate
        :param int udp_port:
            the UDP port on the chip to connect to; connecting to a non-SCP
            port will result in a connection that can't easily be configured.
        :rtype: ~spinnman.connections.udp_packet_connections.SDPConnection
        """

    @abstractmethod
    def open_eieio_connection(self, chip_x, chip_y):
        """
        Open a connection to a specific Ethernet-enabled chip for EIEIO.
        Caller will have to arrange for SpiNNaker to pay attention to the
        connection.

        The coordinates will be job-relative.

        :param int chip_x: Ethernet-enabled chip X coordinate
        :param int chip_y: Ethernet-enabled chip Y coordinate
        :rtype: ~spinnman.connections.udp_packet_connections.EIEIOConnection
        """

    @abstractmethod
    def open_eieio_listener(self):
        """
        Open an unbound EIEIO connection. This may be used to communicate with
        any board of the job.

        :rtype: ~spinnman.connections.udp_packet_connections.EIEIOConnection
        """

    @property
    def proxying(self):
        """
        Whether this is a proxying connection. False unless overridden.

        :rtype: bool
        """
        return False

    def make_report(self, filename):
        """
        Asks the controller to make a report of details of allocations.
        By default, this does nothing.
        """



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from spinn_utilities.config_holder import get_config_int, get_config_str
from spinn_utilities.log import FormatAdapter
from spinn_machine import json_machine, virtual_machine, Machine
logger = FormatAdapter(logging.getLogger(__name__))


def virtual_machine_generator():
    """
    Generates a virtual machine with given dimensions and configuration.

    :return: The virtual machine.
    :rtype: ~spinn_machine.Machine
    :raises Exception: If given bad arguments
    """
    height = get_config_int("Machine", "height")
    width = get_config_int("Machine", "width")
    json_path = get_config_str("Machine", "json_path")

    # For backward compatibility support version in csf files for now
    version = get_config_int("Machine", "version")
    if version is not None:
        if version in [2, 3]:
            if height is None:
                height = 2
            else:
                assert height == 2
            if width is None:
                width = 2
            else:
                assert width == 2
            logger.warning("For virtual Machines version is deprecated."
                           "use width=2, height=2 instead")
        elif version in [4, 5]:
            if height is None:
                height = 8
            else:
                assert height == 8
            if width is None:
                width = 8
            else:
                assert width == 8
            logger.warning("For virtual Machines version is deprecated."
                           "use width=8, height=8 instead")
        else:
            raise ValueError(f"Unknown version {version}")

    if json_path is None:
        machine = virtual_machine(
            width=width, height=height,
            n_cpus_per_chip=Machine.max_cores_per_chip(),
            validate=True)
    else:
        if (height is not None or width is not None or
                version is not None or
                get_config_str("Machine", "down_chips") is not None or
                get_config_str("Machine", "down_cores") is not None or
                get_config_str("Machine", "down_links") is not None):
            logger.warning("As json_path specified all other virtual "
                           "machine settings ignored.")
        machine = json_machine.machine_from_json(json_path)

    # Work out and add the SpiNNaker links and FPGA links
    machine.add_spinnaker_links()
    machine.add_fpga_links()

    logger.info(
        "Created a virtual machine which has {}",
        machine.cores_and_link_output_string())

    return machine



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB
from spinnman.spalloc import SpallocClient


class DatabaseReader(SQLiteDB):
    """
    A reader for the database.
    """
    __slots__ = ("__job", "__looked_for_job")

    def __init__(self, database_path):
        """
        :param str database_path: The path to the database
        """
        super().__init__(database_path, read_only=True, text_factory=str)
        self.__job = None
        self.__looked_for_job = False

    def __exec_one(self, query, *args):
        with self.transaction() as cur:
            cur.execute(query + " LIMIT 1", args)
            return cur.fetchone()

    @staticmethod
    def __r2t(row, *args):
        return tuple(None if row is None else row[key] for key in args)

    def get_job(self):
        """
        Get the job described in the database. If no job exists, direct
        connection to boards should be used.

        :return: Job handle, if one exists. ``None`` otherwise.
        :rtype: ~spinnman.spalloc.SpallocJob
        """
        # This is maintaining an object we only make once
        if not self.__looked_for_job:
            with self.transaction() as cur:
                self.__job = SpallocClient.open_job_from_database(cur)
            self.__looked_for_job = True
        return self.__job

    def get_key_to_atom_id_mapping(self, label):
        """
        Get a mapping of event key to atom ID for a given vertex.

        :param str label: The label of the vertex
        :return: dictionary of atom IDs indexed by event key
        :rtype: dict(int, int)
        """
        with self.transaction() as cur:
            return {
                row["event"]: row["atom"]
                for row in cur.execute(
                    """
                    SELECT * FROM label_event_atom_view
                    WHERE label = ?
                    """, (label, ))}

    def get_atom_id_to_key_mapping(self, label):
        """
        Get a mapping of atom ID to event key for a given vertex.

        :param str label: The label of the vertex
        :return: dictionary of event keys indexed by atom ID
        :rtype: dict(int, int)
        """
        with self.transaction() as cur:
            return {
                row["atom"]: row["event"]
                for row in cur.execute(
                    """
                    SELECT * FROM label_event_atom_view
                    WHERE label = ?
                    """, (label, ))}

    def get_live_output_details(self, label, receiver_label):
        """
        Get the IP address, port and whether the SDP headers are to be
        stripped from the output from a vertex.

        :param str label: The label of the vertex
        :return: tuple of (IP address, port, strip SDP, board address, tag,
            chip_x, chip_y)
        :rtype: tuple(str, int, bool, str, int, int, int)
        """
        return self.__r2t(
            self.__exec_one(
                """
                SELECT * FROM app_output_tag_view
                WHERE pre_vertex_label = ?
                    AND post_vertex_label LIKE ?
                """, label, str(receiver_label) + "%"),
            "ip_address", "port", "strip_sdp", "board_address", "tag",
            "chip_x", "chip_y")

    def get_configuration_parameter_value(self, parameter_name):
        """
        Get the value of a configuration parameter.

        :param str parameter_name: The name of the parameter
        :return: The value of the parameter
        :rtype: float or None
        """
        row = self.__exec_one(
            """
            SELECT value FROM configuration_parameters
            WHERE parameter_id = ?
            """, parameter_name)
        return None if row is None else float(row["value"])

    @staticmethod
    def __xyp(row):
        return int(row["x"]), int(row["y"]), int(row["p"])

    def get_placements(self, label):
        """
        Get the placements of an application vertex with a given label.

        :param str label: The label of the vertex
        :return: A list of x, y, p coordinates of the vertices
        :rtype: list(tuple(int, int, int))
        """
        with self.transaction() as cur:
            return [
                self.__xyp(row) for row in cur.execute(
                    """
                    SELECT x, y, p FROM application_vertex_placements
                    WHERE vertex_label = ?
                    """, (label, ))]

    def get_ip_address(self, x, y):
        """
        Get an IP address to contact a chip.

        :param int x: The x-coordinate of the chip
        :param int y: The y-coordinate of the chip
        :return: The IP address of the Ethernet to use to contact the chip
        :rtype: str or None
        """
        row = self.__exec_one(
            """
            SELECT eth_ip_address FROM chip_eth_info
            WHERE x = ? AND y = ? OR x = 0 AND y = 0
            ORDER BY x DESC
            """, x, y)
        # Should only fail if no machine is present!
        return None if row is None else row["eth_ip_address"]



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractproperty


class AbstractSendMeMulticastCommandsVertex(object, metaclass=AbstractBase):
    """
    A device that may be a virtual vertex which wants to commands to be
    sent to it as multicast packets at fixed points in the simulation.

    .. note::
        The device might not be a vertex at all. It could instead be
        instantiated entirely host side, in which case these methods will
        never be called.
    """

    __slots__ = ()

    @abstractproperty
    def start_resume_commands(self):
        """
        The commands needed when starting or resuming simulation.

        :rtype:
            iterable(~spinn_front_end_common.utility_models.MultiCastCommand)
        """

    @abstractproperty
    def pause_stop_commands(self):
        """
        The commands needed when pausing or stopping simulation.

        :rtype:
            iterable(~spinn_front_end_common.utility_models.MultiCastCommand)
        """

    @abstractproperty
    def timed_commands(self):
        """
        The commands to be sent at given times in the simulation.

        :rtype:
            iterable(~spinn_front_end_common.utility_models.MultiCastCommand)
        """



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinnman.processes import AbstractMultiConnectionProcess
from spinn_front_end_common.utilities.utility_objs.extra_monitor_scp_messages\
    import (
        SetReinjectionPacketTypesMessage)


class SetPacketTypesProcess(AbstractMultiConnectionProcess):
    """
    How to send messages to control what messages are reinjected.
    """

    def set_packet_types(self, core_subsets, point_to_point, multicast,
                         nearest_neighbour, fixed_route):
        """
        Set what types of packets should be reinjected.

        :param ~spinn_machine.CoreSubsets core_subsets:
            sets of cores to send command to
        :param bool point_to_point: If point-to-point should be set
        :param bool multicast: If multicast should be set
        :param bool nearest_neighbour: If nearest neighbour should be set
        :param bool fixed_route: If fixed route should be set
        """
        # pylint: disable=too-many-arguments
        for core_subset in core_subsets.core_subsets:
            for processor_id in core_subset.processor_ids:
                self._send_request(SetReinjectionPacketTypesMessage(
                    core_subset.x, core_subset.y, processor_id, multicast,
                    point_to_point, fixed_route, nearest_neighbour))
        self._finish()
        self.check_for_error()
